from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__docformat__ = 'restructuredtext'

from fipy.tools.numerix import random
from fipy.variables.noiseVariable import NoiseVariable

__all__ = ["BetaNoiseVariable"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class BetaNoiseVariable(NoiseVariable):
    r"""
    Represents a beta distribution of random numbers with the probability
    distribution

    .. math::

       x^{\alpha - 1}\frac{\beta^\alpha e^{-\beta x}}{\Gamma(\alpha)}

    with a shape parameter :math:`\alpha`, a rate parameter :math:`\beta`, and
    :math:`\Gamma(z) = \int_0^\infty t^{z - 1}e^{-t}\,dt`.

    Seed the random module for the sake of deterministic test results.

    >>> from fipy import numerix
    >>> numerix.random.seed(1)

    We generate noise on a uniform Cartesian mesh

    >>> from fipy.variables.variable import Variable
    >>> alpha = Variable()
    >>> beta = Variable()
    >>> from fipy.meshes import Grid2D
    >>> noise = BetaNoiseVariable(mesh = Grid2D(nx = 100, ny = 100), alpha = alpha, beta = beta)

    We histogram the root-volume-weighted noise distribution

    >>> from fipy.variables.histogramVariable import HistogramVariable
    >>> histogram = HistogramVariable(distribution = noise, dx = 0.01, nx = 100)

    and compare to a Gaussian distribution

    >>> from fipy.variables.cellVariable import CellVariable
    >>> betadist = CellVariable(mesh = histogram.mesh)
    >>> x = CellVariable(mesh=histogram.mesh, value=histogram.mesh.cellCenters[0])
    >>> from scipy.special import gamma as Gamma # doctest: +SCIPY
    >>> betadist = ((Gamma(alpha + beta) / (Gamma(alpha) * Gamma(beta)))
    ...             * x**(alpha - 1) * (1 - x)**(beta - 1)) # doctest: +SCIPY

    >>> if __name__ == '__main__':
    ...     from fipy import Viewer
    ...     viewer = Viewer(vars=noise, datamin=0, datamax=1)
    ...     histoplot = Viewer(vars=(histogram, betadist),
    ...                        datamin=0, datamax=1.5)

    >>> from fipy.tools.numerix import arange

    >>> for a in arange(0.5, 5, 0.5):
    ...     alpha.value = a
    ...     for b in arange(0.5, 5, 0.5):
    ...         beta.value = b
    ...         if __name__ == '__main__':
    ...             import sys
    ...             print("alpha: %g, beta: %g" % (alpha, beta), file=sys.stderr)
    ...             viewer.plot()
    ...             histoplot.plot()

    >>> print(abs(noise.faceGrad.divergence.cellVolumeAverage) < 5e-15)
    1

    .. image:: fipy/variables/beta.*
      :scale: 25
      :align: center
      :alt: random values with a beta distribution

    .. image:: fipy/variables/beta-histogram.*
      :scale: 25
      :align: center
      :alt: histogram of random values with a beta distribution

    """
    def __init__(self, mesh, alpha, beta, name = '', hasOld = 0):
        r"""
        Parameters
        ----------
        mesh : ~fipy.meshes.mesh.Mesh
            The mesh on which to define the noise.
        alpha : float
            The parameter :math:`\alpha`.
        beta : float
            The parameter :math:`\beta`.
        """
        NoiseVariable.__init__(self, mesh = mesh, name = name, hasOld = hasOld)
        self.alpha = self._requires(alpha)
        self.beta = self._requires(beta)

    def random(self):
        return random.beta(a = self.alpha, b = self.beta,
                           size = [self.mesh.globalNumberOfCells])

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





from __future__ import unicode_literals
__docformat__ = "restructuredtext"

from fipy.tools import numerix
from fipy.tools import vector
from fipy.tools.dimensions.physicalField import PhysicalField

from fipy.meshes.mesh2D import Mesh2D
from fipy.meshes.representations.gridRepresentation import _Grid2DRepresentation
from fipy.meshes.topologies.meshTopology import _Mesh2DTopology

__all__ = ["Tri2D"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class Tri2D(Mesh2D):
    """
    This class creates a mesh made out of triangles.  It does this by
    starting with a standard Cartesian mesh (`Grid2D`) and dividing each cell
    in that mesh (hereafter referred to as a "box") into four equal
    parts with the dividing lines being the diagonals.
    """

    def __init__(self, dx = 1., dy = 1., nx = 1, ny = 1,
                 _RepresentationClass=_Grid2DRepresentation, _TopologyClass=_Mesh2DTopology):
        """
        Creates a 2D triangular mesh with horizontal faces numbered first then
        vertical faces, then diagonal faces.  Vertices are numbered starting
        with the vertices at the corners of boxes and then the vertices at the
        centers of boxes.  Cells on the right of boxes are numbered first, then
        cells on the top of boxes, then cells on the left of boxes, then cells
        on the bottom of boxes.  Within each of the "sub-categories" in the
        above, the vertices, cells and faces are numbered in the usual way.

        Parameters
        ----------
        float dx, dy : float
            The X and Y dimensions of each "box".
            If `dx` <> `dy`, the line segments connecting the cell
            centers will not be orthogonal to the faces.
        int nx, ny : int
            The number of boxes in the X direction and the Y direction.
            The total number of boxes will be equal to `nx * ny`, and the total
            number of cells will be equal to `4 * nx * ny`.
        """

        self.args = {
            'dx': dx,
            'dy': dy,
            'nx': nx,
            'ny': ny
        }

        self.nx = nx
        self.ny = ny

        self.numberOfHorizontalFaces   = self.nx * (self.ny + 1)
        self.numberOfVerticalFaces     = self.ny * (self.nx + 1)
        self.numberOfEachDiagonalFaces = self.nx * self.ny

        self.dx  = PhysicalField(value = dx)
        scale    = PhysicalField(value = 1, unit = self.dx.unit)
        self.dx /= scale

        self.dy = PhysicalField(value = dy)
        if self.dy.unit.isDimensionless():
            self.dy = dy
        else:
            self.dy /= scale

        self.numberOfCornerVertices = (self.nx + 1) * (self. ny + 1)
        self.numberOfCenterVertices = self.nx * self.ny
        self.numberOfTotalVertices  = self.numberOfCornerVertices + self.numberOfCenterVertices

        self.offset = (0, 0)

        vertices = self._createVertices()
        faces    = self._createFaces()

        cells = self._createCells()
        cells = numerix.sort(cells, axis=0)

        Mesh2D.__init__(self, vertices, faces, cells,
                        _RepresentationClass=_RepresentationClass, _TopologyClass=_TopologyClass)

        self.scale = scale

    def _createVertices(self):

        x = numerix.arange(self.nx + 1) * self.dx
        y = numerix.arange(self.ny + 1) * self.dy
        x = numerix.resize(x, (self.numberOfCornerVertices,))
        y = numerix.repeat(y, self.nx + 1)
        boxCorners = numerix.array((x, y))
        x = numerix.arange(0.5, self.nx + 0.5) * self.dx
        y = numerix.arange(0.5, self.ny + 0.5) * self.dy
        x = numerix.resize(x, (self.numberOfCenterVertices,))
        y = numerix.repeat(y, self.nx)
        boxCenters = numerix.array((x, y))
        return numerix.concatenate((boxCorners, boxCenters), axis=1)

    def _createFaces(self):
        ## v1, v2 refer to the cells.
        ## Horizontal faces are first
        v1 = numerix.arange(self.numberOfCornerVertices)
        v2 = v1 + 1
        horizontalFaces = vector.prune(numerix.array((v1, v2)), self.nx + 1, self.nx, axis=1)
        v1 = numerix.arange(self.numberOfCornerVertices - (self.nx + 1))
        v2 = v1 + self.nx + 1
        verticalFaces =  numerix.array((v1, v2))

        ## reverse some of the face orientations to obtain the correct normals

        tmp = horizontalFaces.copy()
        horizontalFaces[0, :self.nx] = tmp[1, :self.nx]
        horizontalFaces[1, :self.nx] = tmp[0, :self.nx]

        tmp = verticalFaces.copy()
        verticalFaces[0] = tmp[1]
        verticalFaces[1] = tmp[0]
        verticalFaces[0, ::(self.nx + 1)] = tmp[0, ::(self.nx + 1)]
        verticalFaces[1, ::(self.nx + 1)] = tmp[1, ::(self.nx + 1)]

        ## do the center ones now

        cellCenters = numerix.arange(self.numberOfCornerVertices, self.numberOfTotalVertices)
        lowerLefts = vector.prune(numerix.arange(self.numberOfCornerVertices - (self.nx + 1)), self.nx + 1, self.nx)
        lowerRights = lowerLefts + 1
        upperLefts = lowerLefts + self.nx + 1
        upperRights = lowerLefts + self.nx + 2
        lowerLeftFaces = numerix.array((cellCenters, lowerLefts))
        lowerRightFaces = numerix.array((lowerRights, cellCenters))
        upperLeftFaces = numerix.array((cellCenters, upperLefts))
        upperRightFaces = numerix.array((cellCenters, upperRights))
        return numerix.concatenate((horizontalFaces, verticalFaces, lowerLeftFaces, lowerRightFaces, upperLeftFaces, upperRightFaces), axis=1)

    def _createCells(self):
        ## cells = (f1, f2, f3, f4) going anticlockwise.
        ## f1 etc. refer to the faces
        bottomFaces = numerix.arange(0, self.numberOfHorizontalFaces - self.nx)
        topFaces = numerix.arange(self.nx, self.numberOfHorizontalFaces)
        leftFaces = vector.prune(numerix.arange(self.numberOfHorizontalFaces, self.numberOfHorizontalFaces + self.numberOfVerticalFaces), self.nx + 1, self.nx)
        rightFaces = vector.prune(numerix.arange(self.numberOfHorizontalFaces, self.numberOfHorizontalFaces + self.numberOfVerticalFaces), self.nx + 1, 0)
        lowerLeftDiagonalFaces = numerix.arange(self.numberOfHorizontalFaces + self.numberOfVerticalFaces, self.numberOfHorizontalFaces + self.numberOfVerticalFaces + self.numberOfEachDiagonalFaces)
        lowerRightDiagonalFaces = lowerLeftDiagonalFaces + self.numberOfEachDiagonalFaces
        upperLeftDiagonalFaces = lowerRightDiagonalFaces + self.numberOfEachDiagonalFaces
        upperRightDiagonalFaces = upperLeftDiagonalFaces + self.numberOfEachDiagonalFaces
        ##faces in arrays, now get the cells
        bottomOfBoxCells = numerix.array([bottomFaces, lowerRightDiagonalFaces, lowerLeftDiagonalFaces])
        rightOfBoxCells = numerix.array([rightFaces, upperRightDiagonalFaces, lowerRightDiagonalFaces])
        topOfBoxCells = numerix.array([topFaces, upperLeftDiagonalFaces, upperRightDiagonalFaces])
        leftOfBoxCells = numerix.array([leftFaces, lowerLeftDiagonalFaces, upperLeftDiagonalFaces])
        return numerix.concatenate((rightOfBoxCells, topOfBoxCells, leftOfBoxCells, bottomOfBoxCells), axis=1)

    @property
    def physicalShape(self):
        """Return physical dimensions of `Grid2D`.
        """
        return PhysicalField(value = (self.nx * self.dx * self.scale,
                                      self.ny * self.dy * self.scale))

    @property
    def _meshSpacing(self):
        return numerix.array((self.dx, self.dy))[..., numerix.newaxis]

    @property
    def shape(self):
        return (self.nx, self.ny)

    @property
    def _isOrthogonal(self):
        return True

## pickling

    def _test(self):
        """
        These tests are not useful as documentation, but are here to ensure
        everything works as expected.

            >>> dx = 0.5
            >>> dy = 2.
            >>> nx = 3
            >>> ny = 2

            >>> mesh = Tri2D(nx = nx, ny = ny, dx = dx, dy = dy)

            >>> vertices = numerix.array(((0.0, 0.5, 1.0, 1.5, 0.0, 0.5, 1.0, 1.5, 0.0, 0.5, 1.0, 1.5, 0.25, 0.75, 1.25, 0.25, 0.75, 1.25),
            ...                           (0.0, 0.0, 0.0, 0.0, 2.0, 2.0, 2.0, 2.0, 4.0, 4.0, 4.0, 4.0, 1.0,  1.0,  1.0,  3.0,  3.0,  3.0)))

            >>> from fipy.tools import numerix
            >>> numerix.allequal(vertices, mesh._createVertices())
            1

            >>> faces = numerix.array(((1, 2, 3, 4, 5, 6, 8,  9, 10, 0, 5, 6, 7, 4, 9, 10, 11, 12, 13, 14, 15, 16, 17,  1,  2,  3,  5,  6,  7, 12, 13, 14, 15, 16, 17, 12, 13, 14, 15, 16, 17),
            ...                        (0, 1, 2, 5, 6, 7, 9, 10, 11, 4, 1, 2, 3, 8, 5,  6,  7,  0,  1,  2,  4,  5,  6, 12, 13, 14, 15, 16, 17,  4,  5,  6,  8,  9, 10,  5,  6,  7,  9, 10, 11)))
            >>> numerix.allequal(faces, mesh._createFaces())
            1

            >>> cells = numerix.array(((10, 11, 12, 14, 15, 16,  3,  4,  5,  6,  7,  8,  9, 10, 11, 13, 14, 15,  0,  1,  2,  3,  4,  5),
            ...                        (35, 36, 37, 38, 39, 40, 29, 30, 31, 32, 33, 34, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28),
            ...                        (23, 24, 25, 26, 27, 28, 35, 36, 37, 38, 39, 40, 29, 30, 31, 32, 33, 34, 17, 18, 19, 20, 21, 22)))
            >>> numerix.allequal(cells, mesh._createCells())
            1

            >>> externalFaces = numerix.array((0, 1, 2, 6, 7, 8, 9, 12, 13, 16))
            >>> print(numerix.allequal(externalFaces,
            ...                        numerix.nonzero(mesh.exteriorFaces)))
            1

            >>> internalFaces = numerix.array((3, 4, 5, 10, 11, 14, 15, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40))
            >>> print(numerix.allequal(internalFaces,
            ...                        numerix.nonzero(mesh.interiorFaces)))
            1

            >>> from fipy.tools.numerix import MA
            >>> faceCellIds = MA.masked_values(((18, 19, 20,  6,  7,  8,  9, 10, 11, 12,  0,  1,  2, 15,  3,  4,  5, 12, 13, 14, 15, 16, 17,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11,  0,  1,  2,  3,  4,  5),
            ...                                 (-1, -1, -1, 21, 22, 23, -1, -1, -1, -1, 13, 14, -1, -1, 16, 17, -1, 18, 19, 20, 21, 22, 23, 18, 19, 20, 21, 22, 23, 12, 13, 14, 15, 16, 17,  6,  7,  8,  9, 10, 11)), -1)
            >>> numerix.allequal(faceCellIds, mesh.faceCellIDs)
            1

            >>> d = (numerix.sqrt((dx*dx)+(dy*dy))) / 2.0 ## length of diagonal edges
            >>> faceAreas = numerix.array((0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5,
            ...                            2, 2, 2, 2, 2, 2, 2, 2,
            ...                            d, d, d, d, d, d, d, d, d, d, d, d, d, d, d, d, d, d, d, d, d, d, d, d))
            >>> numerix.allclose(faceAreas, mesh._faceAreas, atol = 1e-10, rtol = 1e-10)
            1

            >>> faceCoords = numerix.take(vertices, faces, axis=1)
            >>> faceCenters = (faceCoords[..., 0,:] + faceCoords[..., 1,:]) / 2.
            >>> print(numerix.allclose(faceCenters, mesh.faceCenters, atol = 1e-10, rtol = 1e-10))
            True

            >>> xc = dy  / numerix.sqrt((dx * dx) + (dy * dy))
            >>> yc = dx  / numerix.sqrt((dx * dx) + (dy * dy))

            >>> faceNormals = numerix.array((( 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 1.0, 1.0, 1.0, -1.0, 1.0, 1.0, 1.0, xc, xc, xc, xc, xc, xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc),
            ...                              (-1.0, -1.0, -1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, yc, yc, yc, yc, yc, yc)))
            >>> numerix.allclose(faceNormals, mesh.faceNormals, atol = 1e-10, rtol = 1e-10)
            1

            >>> cellToFaceOrientations = numerix.array((( 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1, 1, -1, -1, 1, 1, 1, -1, -1, -1),
            ...                                         ( 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1, -1, -1, -1, -1),
            ...                                         ( 1, 1, 1, 1, 1, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1)))
            >>> numerix.allequal(cellToFaceOrientations, mesh._cellToFaceOrientations)
            1

            >>> cellVolumes = numerix.array((0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25,
            ...                              0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25,
            ...                              0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25))
            >>> numerix.allclose(cellVolumes, mesh.cellVolumes, atol = 1e-10, rtol = 1e-10)
            1

            >>> sixth = 1.0 / 6.0
            >>> cellCenters = numerix.array(((  5*sixth, 11*sixth, 17*sixth,  5*sixth, 11*sixth, 17*sixth,      0.5,      1.5,      2.5,      0.5,      1.5,      2.5,  1*sixth,  7*sixth, 13*sixth,  1*sixth,  7*sixth, 13*sixth,      0.5,      1.5,      2.5,      0.5,      1.5,      2.5),
            ...                              (      0.5,      0.5,      0.5,      1.5,      1.5,      1.5,  5*sixth,  5*sixth,  5*sixth, 11*sixth, 11*sixth, 11*sixth,      0.5,      0.5,      0.5,      1.5,      1.5,      1.5,  1*sixth,  1*sixth,  1*sixth,  7*sixth,  7*sixth,  7*sixth)))
            >>> cellCenters *= numerix.array([[dx], [dy]])
            >>> print(numerix.allclose(cellCenters, mesh.cellCenters, atol = 1e-10, rtol = 1e-10))
            True

            >>> yd = numerix.sqrt(((dx/12.0)*(dx/12.0)) + ((dy/ 4.0)*(dy/ 4.0)))
            >>> xd = numerix.sqrt(((dx/ 4.0)*(dx/ 4.0)) + ((dy/12.0)*(dy/12.0)))
            >>> faceToCellDistances = MA.masked_values(((dy/6.0, dy/6.0, dy/6.0, dy/6.0, dy/6.0, dy/6.0,  dy/6.0, dy/6.0, dy/6.0, dx/6.0, dx/6.0, dx/6.0, dx/6.0, dx/6.0, dx/6.0, dx/6.0, dx/6.0,     yd,     yd,     yd,     yd,     yd,     yd,     yd,     yd,     yd,     yd,     yd,     yd,     xd,     xd,     xd,     xd,     xd,     xd,     yd,     yd,     yd,     yd,     yd,     yd),
            ...                                         (    -1,     -1,     -1, dy/6.0, dy/6.0, dy/6.0,      -1,     -1,     -1,     -1, dx/6.0, dx/6.0,     -1,     -1, dx/6.0, dx/6.0,     -1,     xd,     xd,     xd,     xd,     xd,     xd,     xd,     xd,     xd,     xd,     xd,     xd,     yd,     yd,     yd,     yd,     yd,     yd,     xd,     xd,     xd,     xd,     xd,     xd)), -1)
            >>> numerix.allclose(faceToCellDistances, mesh._faceToCellDistances, atol = 1e-10, rtol = 1e-10)
            1

            >>> dd = numerix.sqrt((dx*dx)+(dy*dy)) / 3.0
            >>> cellDistances = numerix.array((dy/6.0, dy/6.0, dy/6.0, dy/3.0, dy/3.0, dy/3.0, dy/6.0, dy/6.0, dy/6.0,
            ...                                dx/6.0, dx/3.0, dx/3.0, dx/6.0, dx/6.0, dx/3.0, dx/3.0, dx/6.0,
            ...                                dd, dd, dd, dd, dd, dd, dd, dd, dd, dd, dd, dd,
            ...                                dd, dd, dd, dd, dd, dd, dd, dd, dd, dd, dd, dd))
            >>> numerix.allclose(cellDistances, mesh._cellDistances, atol = 1e-10, rtol = 1e-10)
            1

            >>> faceToCellDistanceRatios = faceToCellDistances[0] / cellDistances
            >>> numerix.allclose(faceToCellDistanceRatios, mesh._faceToCellDistanceRatio, atol = 1e-10, rtol = 1e-10)
            1

            >>> areaProjections = faceNormals * faceAreas
            >>> numerix.allclose(areaProjections, mesh._areaProjections, atol = 1e-10, rtol = 1e-10)
            1

            >>> tangents1 = numerix.array(((1.0, 1.0, 1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,  yc,  yc,  yc,  yc,  yc,  yc,  yc,  yc,  yc,  yc,  yc,  yc,  yc,  yc,  yc,  yc,  yc,  yc, -yc, -yc, -yc, -yc, -yc, -yc),
            ...                            (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 1.0, 1.0, 1.0, -1.0, 1.0, 1.0, 1.0,  xc,  xc,  xc,  xc,  xc,  xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc)))
            >>> numerix.allclose(tangents1, mesh._faceTangents1, atol = 1e-10, rtol = 1e-10)
            1

            >>> tangents2 = numerix.zeros((2, 41), 'l')
            >>> numerix.allclose(tangents2, mesh._faceTangents2, atol = 1e-10, rtol = 1e-10)
            1

            >>> cellToCellIDs = MA.masked_values(((13, 14, -1, 16, 17, -1, 21, 22, 23, -1, -1, -1, -1,  0,  1, -1,  3,  4, -1, -1, -1,  6,  7,  8),
            ...                                   (18, 19, 20, 21, 22, 23, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 12, 13, 14, 15, 16, 17),
            ...                                   ( 6,  7,  8,  9, 10, 11,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11,  0,  1,  2,  3,  4,  5)), -1)
            >>> numerix.allequal(cellToCellIDs, mesh._cellToCellIDs)
            1

            >>> cellToCellDistances = numerix.array(((dx/3.0, dx/3.0, dx/6.0, dx/3.0, dx/3.0, dx/6.0, dy/3.0, dy/3.0, dy/3.0, dy/6.0, dy/6.0, dy/6.0, dx/6.0, dx/3.0, dx/3.0, dx/6.0, dx/3.0, dx/3.0, dy/6.0, dy/6.0, dy/6.0, dy/3.0, dy/3.0, dy/3.0),
            ...                                      (    dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd),
            ...                                      (    dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd,     dd)))
            >>> numerix.allclose(cellToCellDistances, mesh._cellToCellDistances, atol = 1e-10, rtol = 1e-10)
            1

            >>> interiorCellIDs = numerix.array((0, 1, 3, 4, 6, 7, 8, 13, 14, 16, 17, 21, 22, 23))
            >>> numerix.allequal(interiorCellIDs, mesh._interiorCellIDs)
            1

            >>> exteriorCellIDs = numerix.array((2, 5, 9, 10, 11, 12, 15, 18, 19, 20))
            >>> numerix.allequal(exteriorCellIDs, mesh._exteriorCellIDs)
            1

            >>> cellNormals = numerix.array((((  1,  1,  1,  1,  1,  1,  0,  0,  0,  0,  0,  0, -1, -1, -1, -1, -1, -1,  0,  0,  0,  0,  0,  0),
            ...                               (-xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, -xc, xc, xc, xc, xc, xc, xc, -xc, -xc, -xc, -xc, -xc, -xc),
            ...                               (-xc, -xc, -xc, -xc, -xc, -xc, xc, xc, xc, xc, xc, xc, xc, xc, xc, xc, xc, xc, xc, xc, xc, xc, xc, xc)),
            ...                              ((  0,  0,  0,  0,  0,  0,  1,  1,  1,  1,  1,  1,  0,  0,  0,  0,  0,  0, -1, -1, -1, -1, -1, -1),
            ...                               (-yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, -yc, yc, yc, yc, yc, yc, yc),
            ...                               ( yc, yc, yc, yc, yc, yc, -yc, -yc, -yc, -yc, -yc, -yc, yc, yc, yc, yc, yc, yc, yc, yc, yc, yc, yc, yc))))
            >>> numerix.allclose(cellNormals, mesh._cellNormals, atol = 1e-10, rtol = 1e-10)
            1

            >>> cellAreaProjections = numerix.array((((     dy,     dy,     dy,     dy,     dy,     dy,      0,      0,      0,      0,      0,      0,    -dy,    -dy,    -dy,    -dy,    -dy,    -dy,      0,      0,      0,      0,      0,      0),
            ...                                       ( -dy/2., -dy/2., -dy/2., -dy/2., -dy/2., -dy/2., -dy/2., -dy/2., -dy/2., -dy/2., -dy/2., -dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2., -dy/2., -dy/2., -dy/2., -dy/2., -dy/2., -dy/2.),
            ...                                       ( -dy/2., -dy/2., -dy/2., -dy/2., -dy/2., -dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.,  dy/2.)),
            ...                                      ((      0,      0,      0,      0,      0,      0,     dx,     dx,     dx,     dx,     dx,     dx,      0,      0,      0,      0,      0,      0,    -dx,    -dx,    -dx,    -dx,    -dx,    -dx),
            ...                                       ( -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2.),
            ...                                       (  dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2., -dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2.,  dx/2.))))
            >>> numerix.allclose(cellAreaProjections, mesh._cellAreaProjections, atol = 1e-10, rtol = 1e-10)
            1

            >>> tmp1 = numerix.array((12, 5, 1))
            >>> tmp2 = numerix.array((12, 5, 4))
            >>> tmp3 = numerix.array((12, 4, 0))
            >>> tmp4 = numerix.array((12, 1, 0))
            >>> tmp5 = numerix.array((0, 1, 1))
            >>> cellVertexIDs = numerix.array((tmp1, tmp1 + 1, tmp1 + 2, tmp1 + 3 + tmp5, tmp1 + 4 + tmp5, tmp1 + 5 + tmp5,
            ...                                tmp2, tmp2 + 1, tmp2 + 2, tmp2 + 3 + tmp5, tmp2 + 4 + tmp5, tmp2 + 5 + tmp5,
            ...                                tmp3, tmp3 + 1, tmp3 + 2, tmp3 + 3 + tmp5, tmp3 + 4 + tmp5, tmp3 + 5 + tmp5,
            ...                                tmp4, tmp4 + 1, tmp4 + 2, tmp4 + 3 + tmp5, tmp4 + 4 + tmp5, tmp4 + 5 + tmp5))
            >>> cellVertexIDs = cellVertexIDs.swapaxes(0, 1)
            >>> numerix.allclose(mesh._cellVertexIDs, cellVertexIDs)
            1


            >>> from fipy.tools import dump
            >>> (f, filename) = dump.write(mesh, extension = '.gz')
            >>> unpickledMesh = dump.read(filename, f)

            >>> print(numerix.allequal(mesh.cellCenters, unpickledMesh.cellCenters))
            True
        """

## test test test

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





from __future__ import division
from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.terms.faceTerm import FaceTerm
from fipy.variables.meshVariable import _MeshVariable
from fipy.variables.faceVariable import FaceVariable
from fipy.variables.cellVariable import CellVariable
from fipy.terms import AbstractBaseClassError
from fipy.terms import VectorCoeffError
from fipy.tools import numerix

class _AbstractConvectionTerm(FaceTerm):
    """
    .. attention:: This class is abstract. Always create one of its subclasses.
    """
    def __init__(self, coeff=1.0, var=None):
        """
        Create a `_AbstractConvectionTerm` object.

            >>> from fipy import *
            >>> m = Grid1D(nx = 2)
            >>> cv = CellVariable(mesh = m)
            >>> fv = FaceVariable(mesh = m)
            >>> vcv = CellVariable(mesh=m, rank=1)
            >>> vfv = FaceVariable(mesh=m, rank=1)
            >>> __ConvectionTerm(coeff = cv) # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ...
            VectorCoeffError: The coefficient must be a vector value.
            >>> __ConvectionTerm(coeff = fv) # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ...
            VectorCoeffError: The coefficient must be a vector value.
            >>> __ConvectionTerm(coeff = vcv)
            __ConvectionTerm(coeff=_ArithmeticCellToFaceVariable(value=array([[ 0.,  0.,  0.]]), mesh=UniformGrid1D(dx=1.0, nx=2)))
            >>> __ConvectionTerm(coeff = vfv)
            __ConvectionTerm(coeff=FaceVariable(value=array([[ 0.,  0.,  0.]]), mesh=UniformGrid1D(dx=1.0, nx=2)))
            >>> __ConvectionTerm(coeff = (1,))
            __ConvectionTerm(coeff=(1,))
            >>> ExplicitUpwindConvectionTerm(coeff = (0,)).solve(var=cv, solver=DummySolver()) # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ...
            TransientTermError: The equation requires a TransientTerm with explicit convection.
            >>> (TransientTerm(0.) - ExplicitUpwindConvectionTerm(coeff = (0,))).solve(var=cv, solver=DummySolver(), dt=1.)

            >>> (TransientTerm() - ExplicitUpwindConvectionTerm(coeff = 1)).solve(var=cv, solver=DummySolver(), dt=1.) # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ...
            VectorCoeffError: The coefficient must be a vector value.
            >>> m2 = Grid2D(nx=2, ny=1)
            >>> cv2 = CellVariable(mesh=m2)
            >>> vcv2 = CellVariable(mesh=m2, rank=1)
            >>> vfv2 = FaceVariable(mesh=m2, rank=1)
            >>> __ConvectionTerm(coeff=vcv2)
            __ConvectionTerm(coeff=_ArithmeticCellToFaceVariable(value=array([[ 0.,  0.,  0.,  0.,  0.,  0.,  0.],
                   [ 0.,  0.,  0.,  0.,  0.,  0.,  0.]]), mesh=UniformGrid2D(dx=1.0, nx=2, dy=1.0, ny=1)))
            >>> __ConvectionTerm(coeff=vfv2)
            __ConvectionTerm(coeff=FaceVariable(value=array([[ 0.,  0.,  0.,  0.,  0.,  0.,  0.],
                   [ 0.,  0.,  0.,  0.,  0.,  0.,  0.]]), mesh=UniformGrid2D(dx=1.0, nx=2, dy=1.0, ny=1)))
            >>> (TransientTerm() - ExplicitUpwindConvectionTerm(coeff = ((0,), (0,)))).solve(var=cv2, solver=DummySolver(), dt=1.)
            >>> (TransientTerm() - ExplicitUpwindConvectionTerm(coeff = (0, 0))).solve(var=cv2, solver=DummySolver(), dt=1.)


        Parameters
        ----------
        coeff : The `Term`'s coefficient value.
        """
        if self.__class__ is _AbstractConvectionTerm:
            raise AbstractBaseClassError

        self.stencil = None

        if isinstance(coeff, _MeshVariable) and coeff.rank < 1:
            raise VectorCoeffError

        if isinstance(coeff, CellVariable):
            coeff = coeff.arithmeticFaceValue

        FaceTerm.__init__(self, coeff=coeff, var=var)

    def _calcGeomCoeff(self, var):
        mesh = var.mesh

        if not isinstance(self.coeff, FaceVariable):
            shape = numerix.array(self.coeff).shape

            if shape != () and shape != (1,) and shape[-1] == 1:
                shape = shape[:-1]

            self.coeff = FaceVariable(mesh=mesh, elementshape=shape, value=self.coeff)

        projectedCoefficients = self.coeff * mesh._orientedAreaProjections

        return projectedCoefficients.sum(0)

    def _getWeight(self, var, transientGeomCoeff=None, diffusionGeomCoeff=None):
        r"""
        Testing that the sign of the equation is taken into account
        when evaluation upwind direction.

        >>> from fipy import Grid1D, CellVariable
        >>> from fipy import TransientTerm, ConvectionTerm, ImplicitSourceTerm

        >>> m = Grid1D(nx=3, dx=0.5)
        >>> v = CellVariable(mesh=m)
        >>> v.constrain(1, m.facesLeft)
        >>> v.constrain(0, m.facesRight)

        >>> (-TransientTerm(coeff=1 / m.x) ==
        ...  ConvectionTerm(coeff=[[1]])
        ...  + ImplicitSourceTerm(coeff=m.x)).solve(v, dt=1.)

        >>> v0 = v.copy()
        >>> v[:] = 0
        >>> (TransientTerm(coeff=1 / m.x) ==
        ...  - ConvectionTerm(coeff=[[1]])
        ...  - ImplicitSourceTerm(coeff=m.x)).solve(v, dt=1.)

        >>> print(numerix.allclose(v, v0))
        True

        """

        if self.stencil is None:

            geomCoeff = self._getGeomCoeff(var)
            large = 1e+20
            pecletLarge = large - (geomCoeff < 0) * (2 * large)
            if numerix.all(self._getDiagonalSign(transientGeomCoeff, diffusionGeomCoeff) < 0):
                pecletLarge = -pecletLarge

            if diffusionGeomCoeff is None or diffusionGeomCoeff[0] is None:
                peclet = pecletLarge
            else:
                diffCoeff = diffusionGeomCoeff[0].numericValue
                diffCoeff = diffCoeff - (diffCoeff == 0) * geomCoeff / pecletLarge
                peclet = -geomCoeff / diffCoeff

            alpha = self._alpha(peclet)

            self.stencil = {'implicit' : {'cell 1 diag'    : alpha,
                                          'cell 1 offdiag' : (1-alpha),
                                          'cell 2 diag'    : -(1-alpha),
                                          'cell 2 offdiag' : -alpha}}

        return self.stencil

    def _checkVar(self, var):
        FaceTerm._checkVar(self, var)
        if not (isinstance(self.coeff, FaceVariable) and self.coeff.rank == 1):
            coeffShape = numerix.getShape(self.coeff)
            if (coeffShape is ()) or (coeffShape[0] != var.mesh.dim):
                raise VectorCoeffError

    def _buildMatrix(self, var, SparseMatrix, boundaryConditions=(), dt=None, transientGeomCoeff=None, diffusionGeomCoeff=None):

        var, L, b = FaceTerm._buildMatrix(self, var, SparseMatrix, boundaryConditions=boundaryConditions, dt=dt, transientGeomCoeff=transientGeomCoeff, diffusionGeomCoeff=diffusionGeomCoeff)

##        if var.rank != 1:

        mesh = var.mesh

        if (not hasattr(self, 'constraintL')) or (not hasattr(self, 'constraintB')):

            weight = self._getWeight(var, transientGeomCoeff, diffusionGeomCoeff)

            if 'implicit' in weight:
                alpha = weight['implicit']['cell 1 diag']
            else:
                alpha = 0.0

            alpha_constraint = numerix.where(var.faceGrad.constraintMask, 1.0, alpha)

            def divergence(face_value):
                return (
                    face_value * \
                    (var.faceGrad.constraintMask | var.arithmeticFaceValue.constraintMask) * \
                    self.coeff * mesh.exteriorFaces
                ).divergence * mesh.cellVolumes

            self.constraintL = divergence(alpha_constraint)
            dvar = (var.faceGrad * mesh._cellDistances * mesh.faceNormals).sum(axis=0)
            self.constraintB = divergence(
                (alpha_constraint - 1) * var.arithmeticFaceValue + (alpha - 1)  * dvar * var.faceGrad.constraintMask
            )


        ids = self._reshapeIDs(var, numerix.arange(mesh.numberOfCells))
        L.addAt(numerix.array(self.constraintL).ravel(), ids.ravel(), ids.swapaxes(0, 1).ravel())
        b += numerix.reshape(self.constraintB.value, ids.shape).sum(0).ravel()

        return (var, L, b)


    def _test(self):
        """Test cases for convection with constraints.

        The following tests both a Dirichlet and a Neumann type
        boundary condition with convection using constraints. It
        checks that the Neumann boundary condition converges without
        multiple iterations of the solve step. The Dirichlet boundary
        condition is second order accurate. The Neumann is only first
        order accurate. These tests were prompted by
        https://github.com/usnistgov/fipy/pull/714.

        >>> import fipy as fp
        >>> import numpy as np

        Test Dirichlet boundary condition with a fixed value on the
        left and a fixed value on the right. Check that the order of
        accuracy is 2.


        >>> nx0 = 64
        >>> mesh = fp.Grid1D(nx=nx0, dx=1. / nx0)
        >>> var = fp.CellVariable(mesh)
        >>> var.constrain(1.0, mesh.facesLeft)
        >>> var.constrain(0.0, mesh.facesRight)
        >>> eqn = fp.CentralDifferenceConvectionTerm((1.0,)) - fp.DiffusionTerm()
        >>> _ = eqn.sweep(var, solver=fp.LinearLUSolver())
        >>> expected =  (np.exp(mesh.x) - np.exp(1.)) / (1 - np.exp(1.))
        >>> error0 = float(np.sqrt(((var.value - expected)**2 * mesh.dx).sum()))

        >>> nx1 = 128
        >>> mesh = fp.Grid1D(nx=nx1, dx=1. / nx1)
        >>> var = fp.CellVariable(mesh)
        >>> var.constrain(1.0, mesh.facesLeft)
        >>> var.constrain(0.0, mesh.facesRight)
        >>> eqn = fp.CentralDifferenceConvectionTerm((1.0,)) - fp.DiffusionTerm()
        >>> _ = eqn.sweep(var, solver=fp.LinearLUSolver())
        >>> expected =  (np.exp(mesh.x) - np.exp(1.)) / (1 - np.exp(1.))
        >>> error1 = float(np.sqrt(((var.value - expected)**2 * mesh.dx).sum()))

        >>> assert np.allclose(np.log(error1 / error0 ) / np.log(nx0 / nx1), 2.0, atol=0.02)

        Test Neumann boundary condition with a fixed value on the left
        and a gradient on the right. Check that the order of accuracy
        is 1.

        >>> dphi = 1.0

        >>> nx0 = 64
        >>> mesh = fp.Grid1D(nx=nx0, dx=1. / nx0)
        >>> var = fp.CellVariable(mesh)
        >>> var.constrain(1.0, mesh.facesLeft)
        >>> _  = var.faceGrad.constrain([dphi], mesh.facesRight),
        >>> eqn = fp.CentralDifferenceConvectionTerm((1.0,)) - fp.DiffusionTerm()
        >>> _ = eqn.sweep(var, solver=fp.LinearLUSolver())
        >>> expected = 1 + dphi * (np.exp(mesh.x) - 1) / np.exp(1.)
        >>> error0 = float(np.sqrt(((var.value - expected)**2 * mesh.dx).sum()))

        >>> nx1 = 128
        >>> mesh = fp.Grid1D(nx=nx1, dx=1. / nx1)
        >>> var = fp.CellVariable(mesh)
        >>> var.constrain(1.0, mesh.facesLeft)
        >>> _ = var.faceGrad.constrain([dphi], mesh.facesRight),
        >>> eqn = fp.CentralDifferenceConvectionTerm((1.0,)) - fp.DiffusionTerm()
        >>> _ = eqn.sweep(var, solver=fp.LinearLUSolver())
        >>> expected = 1 + dphi * (np.exp(mesh.x) - 1) / np.exp(1.)
        >>> error1 = float(np.sqrt(((var.value - expected)**2 * mesh.dx).sum()))

        >>> assert np.allclose(np.log(error1 / error0 ) / np.log(nx0 / nx1), 1.0, atol=0.002)

        """

class __ConvectionTerm(_AbstractConvectionTerm):
    """
    Dummy subclass for tests
    """
    pass

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



from __future__ import unicode_literals
from builtins import object
from pyamg import smoothed_aggregation_solver

__all__ = ["SmoothedAggregationPreconditioner"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class SmoothedAggregationPreconditioner(object):
    def __init__(self):
        pass

    def _applyToMatrix(self, A):
        return smoothed_aggregation_solver(A).aspreconditioner(cycle='V')



from __future__ import unicode_literals
import numpy
from scipy.sparse import csr_matrix

import pyamgx

from fipy.solvers.solver import Solver
from fipy.matrices.scipyMatrix import _ScipyMeshMatrix
from fipy.tools import numerix

__all__ = ["PyAMGXSolver"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class PyAMGXSolver(Solver):

    def __init__(self, config_dict, tolerance=1e-10, iterations=2000,
                 precon=None, smoother=None, **kwargs):
        """
        Parameters
        ----------
        config_dict : dict
            AMGX configuration options
        tolerance : float
            Required error tolerance.
        iterations : int
            Maximum number of iterative steps to perform.
        precon : ~fipy.solvers.pyamgx.preconditioners.preconditioners.Preconditioner, optional
        smoother : ~fipy.solvers.pyamgx.smoothers.smoothers.Smoother, optional
        **kwargs
            Other AMGX solver options
        """
        # update solver config:
        config_dict["solver"]["tolerance"] = tolerance
        config_dict["solver"]["max_iters"] = iterations
        if precon:
            config_dict["solver"]["preconditioner"] = precon
        if smoother:
            config_dict["solver"]["smoother"] = smoother
        config_dict["solver"].update(kwargs)

        # create AMGX objects:
        self.cfg = pyamgx.Config().create_from_dict(config_dict)
        self.resources = pyamgx.Resources().create_simple(self.cfg)
        self.x_gpu = pyamgx.Vector().create(self.resources)
        self.b_gpu = pyamgx.Vector().create(self.resources)
        self.A_gpu = pyamgx.Matrix().create(self.resources)
        self.solver = pyamgx.Solver().create(self.resources, self.cfg)

        super(PyAMGXSolver, self).__init__(tolerance=tolerance, iterations=iterations)

    def __exit__(self, *args):
        # destroy AMGX objects:
        self.A_gpu.destroy()
        self.b_gpu.destroy()
        self.x_gpu.destroy()
        self.solver.destroy()
        self.resources.destroy()
        self.cfg.destroy()

    @property
    def _matrixClass(self):
        return _ScipyMeshMatrix

    def _storeMatrix(self, var, matrix, RHSvector):
        self.var = var
        self.matrix = matrix
        self.RHSvector = RHSvector
        self.A_gpu.upload_CSR(self.matrix.matrix)
        self.solver.setup(self.A_gpu)

    def _solve_(self, L, x, b):
        # transfer data from CPU to GPU
        self.x_gpu.upload(x)
        self.b_gpu.upload(b)

        # solve system on GPU
        self.solver.solve(self.b_gpu, self.x_gpu)

        # download values from GPU to CPU
        self.x_gpu.download(x)
        return x

    def _solve(self):
         if self.var.mesh.communicator.Nproc > 1:
             raise Exception("SciPy solvers cannot be used with multiple processors")

         self.var[:] = numerix.reshape(self._solve_(self.matrix, self.var.ravel(), numerix.array(self.RHSvector)), self.var.shape)



from __future__ import division
from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.variables.cellVariable import CellVariable
from fipy.tools import numerix

__all__ = ["SurfactantVariable"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class SurfactantVariable(CellVariable):
    """

    The `SurfactantVariable` maintains a conserved volumetric
    concentration on cells adjacent to, but in front of, the
    interface. The `value` argument corresponds to the initial
    concentration of surfactant on the interface (moles divided by
    area). The value held by the `SurfactantVariable` is actually a
    volume density (moles divided by volume).

    """

    def __init__(self, value = 0., distanceVar = None, name = 'surfactant variable', hasOld=False):
        """

        A simple 1D test:

        >>> from fipy.meshes import Grid1D
        >>> mesh = Grid1D(dx = 1., nx = 4)
        >>> from fipy.variables.distanceVariable import DistanceVariable
        >>> distanceVariable = DistanceVariable(mesh = mesh,
        ...                                     value = (-1.5, -0.5, 0.5, 941.5))
        >>> surfactantVariable = SurfactantVariable(value = 1,
        ...                                         distanceVar = distanceVariable)
        >>> print(numerix.allclose(surfactantVariable, (0, 0., 1., 0)))
        1

        A 2D test case:

        >>> from fipy.meshes import Grid2D
        >>> mesh = Grid2D(dx = 1., dy = 1., nx = 3, ny = 3)
        >>> distanceVariable = DistanceVariable(mesh = mesh,
        ...                                     value = (1.5, 0.5, 1.5,
        ...                                              0.5, -0.5, 0.5,
        ...                                              1.5, 0.5, 1.5))
        >>> surfactantVariable = SurfactantVariable(value = 1,
        ...                                         distanceVar = distanceVariable)
        >>> print(numerix.allclose(surfactantVariable, (0, 1, 0, 1, 0, 1, 0, 1, 0)))
        1

        Another 2D test case:

        >>> mesh = Grid2D(dx = .5, dy = .5, nx = 2, ny = 2)
        >>> distanceVariable = DistanceVariable(mesh = mesh,
        ...                                     value = (-0.5, 0.5, 0.5, 1.5))
        >>> surfactantVariable = SurfactantVariable(value = 1,
        ...                                         distanceVar = distanceVariable)
        >>> print(numerix.allclose(surfactantVariable,
        ...                  (0, numerix.sqrt(2), numerix.sqrt(2), 0)))
        1

        Parameters
        ----------
        value : float or array_like
            The initial value.
        distanceVar : ~fipy.variables.distanceVariable.DistanceVariable
        name : str
            The name of the variable.

        """


        CellVariable.__init__(self, mesh = distanceVar.mesh, name = name, hasOld=False)

        self.distanceVar = self._requires(distanceVar)
        self._value = numerix.array(distanceVar.cellInterfaceAreas) * value / self.mesh.cellVolumes

        if hasOld:
            self._old = self.copy()
        else:
            self._old = None

        self.interfaceSurfactantVariable = None

    @property
    def interfaceVar(self):
        """

        Returns the `SurfactantVariable` rendered as an
        `_InterfaceSurfactantVariable` which evaluates the surfactant
        concentration as an area concentration the interface rather
        than a volumetric concentration.

        """
        if self.interfaceSurfactantVariable is None:
            self.interfaceSurfactantVariable = _InterfaceSurfactantVariable(self)

        return self.interfaceSurfactantVariable

    def _calcValue(self):
        return self._value

    def copy(self):
        return self.__class__(
            distanceVar=self.distanceVar,
            name=self.name + "_old",
            value=self.value.copy(),
            hasOld=False)

class _InterfaceSurfactantVariable(CellVariable):
    def __init__(self, surfactantVar):
        CellVariable.__init__(self, name = surfactantVar.name + "_interface", mesh = surfactantVar.mesh)
        self.surfactantVar = self._requires(surfactantVar)

    def _calcValue(self):
        areas = numerix.array(self.surfactantVar.distanceVar.cellInterfaceAreas)
        areas = numerix.where(areas > 1e-20, areas, 1)
        return numerix.array(self.surfactantVar) * self.mesh.cellVolumes / areas

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





from __future__ import division
from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = ["AdvectionTerm"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

from fipy.tools.numerix import MA
from fipy.tools import numerix

from fipy.terms.firstOrderAdvectionTerm import FirstOrderAdvectionTerm

class AdvectionTerm(FirstOrderAdvectionTerm):
    r"""

    The `AdvectionTerm` object constructs the `b` vector contribution for
    the advection term given by

    .. math::

       u \abs{\nabla \phi}

    from the advection equation given by:

    .. math::

       \frac{\partial \phi}{\partial t} + u \abs{\nabla \phi} = 0

    The construction of the gradient magnitude term requires upwinding as in the standard
    `FirstOrderAdvectionTerm`. The higher order terms are incorporated as follows.
    The formula used here is given by:

    .. math::

       u_P \abs{\nabla \phi}_P = \max \left( u_P , 0 \right) \left[  \sum_A \min \left( D_{AP}, 0 \right)^2 \right]^{1/2} +  \min \left( u_P , 0 \right) \left[  \sum_A \max \left( D_{AP}, 0 \right)^2 \right]^{1/2}

    where,

    .. math::

       D_{AP} = \frac{ \phi_A - \phi_P } { d_{AP}} - \frac{ d_{AP} } {2} m \left(L_A, L_P \right)

    and

    .. math::

       m\left(x, y\right) &= x \qquad \text{if $\abs{x} \le \abs{y} \forall xy \ge 0$} \\
       m\left(x, y\right) &= y \qquad \text{if $\abs{x} > \abs{y} \forall xy \ge 0$} \\
       m\left(x, y\right) &= 0 \qquad \text{if $xy < 0$}

    also,

    .. math::

       L_A &= \frac{\phi_{AA} + \phi_P - 2 \phi_A}{d_{AP}^2} \\
       L_P &= \frac{\phi_{A} + \phi_{PP} - 2 \phi_P}{d_{AP}^2}

    Here are some simple test cases for this problem:

    >>> from fipy.meshes import Grid1D
    >>> from fipy.solvers import *
    >>> SparseMatrix = LinearPCGSolver()._matrixClass
    >>> mesh = Grid1D(dx = 1., nx = 3)

    Trivial test:

    >>> from fipy.variables.cellVariable import CellVariable
    >>> coeff = CellVariable(mesh = mesh, value = numerix.zeros(3, 'd'))
    >>> v, L, b = AdvectionTerm(0.)._buildMatrix(coeff, SparseMatrix)
    >>> print(numerix.allclose(b, numerix.zeros(3, 'd'), atol = 1e-10)) # doctest: +PROCESSOR_0
    True

    Less trivial test:

    >>> coeff = CellVariable(mesh = mesh, value = numerix.arange(3))
    >>> v, L, b = AdvectionTerm(1.)._buildMatrix(coeff, SparseMatrix)
    >>> print(numerix.allclose(b, numerix.array((0., -1., -1.)), atol = 1e-10)) # doctest: +PROCESSOR_0
    True

    Even less trivial

    >>> coeff = CellVariable(mesh = mesh, value = numerix.arange(3))
    >>> v, L, b = AdvectionTerm(-1.)._buildMatrix(coeff, SparseMatrix)
    >>> print(numerix.allclose(b, numerix.array((1., 1., 0.)), atol = 1e-10)) # doctest: +PROCESSOR_0
    True

    Another trivial test case (more trivial than a trivial test case
    standing on a harpsichord singing "trivial test cases are here again")

    >>> vel = numerix.array((-1, 2, -3))
    >>> coeff = CellVariable(mesh = mesh, value = numerix.array((4, 6, 1)))
    >>> v, L, b = AdvectionTerm(vel)._buildMatrix(coeff, SparseMatrix)
    >>> print(numerix.allclose(b, -vel * numerix.array((2, numerix.sqrt(5**2 + 2**2), 5)), atol = 1e-10)) # doctest: +PROCESSOR_0
    True

    Somewhat less trivial test case:

    >>> from fipy.meshes import Grid2D
    >>> mesh = Grid2D(dx = 1., dy = 1., nx = 2, ny = 2)
    >>> vel = numerix.array((3, -5, -6, -3))
    >>> coeff = CellVariable(mesh = mesh, value = numerix.array((3, 1, 6, 7)))
    >>> v, L, b = AdvectionTerm(vel)._buildMatrix(coeff, SparseMatrix)
    >>> answer = -vel * numerix.array((2, numerix.sqrt(2**2 + 6**2), 1, 0))
    >>> print(numerix.allclose(b, answer, atol = 1e-10)) # doctest: +PROCESSOR_0
    True

    For the above test cases the `AdvectionTerm` gives the
    same result as the `AdvectionTerm`. The following test imposes a quadratic
    field. The higher order term can resolve this field correctly.

    .. math::

       \phi = x^2

    The returned vector ``b`` should have the value:

    .. math::

       -\abs{\nabla \phi} = -\left|\frac{\partial \phi}{\partial x}\right| = - 2 \abs{x}

    Build the test case in the following way,

    >>> mesh = Grid1D(dx = 1., nx = 5)
    >>> vel = 1.
    >>> coeff = CellVariable(mesh = mesh, value = mesh.cellCenters[0]**2)
    >>> v, L, b = __AdvectionTerm(vel)._buildMatrix(coeff, SparseMatrix)

    The first order term is not accurate. The first and last element are ignored because they
    don't have any neighbors for higher order evaluation

    >>> print(numerix.allclose(CellVariable(mesh=mesh,
    ... value=b).globalValue[1:-1], -2 * mesh.cellCenters.globalValue[0][1:-1]))
    False

    The higher order term is spot on.

    >>> v, L, b = AdvectionTerm(vel)._buildMatrix(coeff, SparseMatrix)
    >>> print(numerix.allclose(CellVariable(mesh=mesh,
    ... value=b).globalValue[1:-1], -2 * mesh.cellCenters.globalValue[0][1:-1]))
    True

    The `AdvectionTerm` will also resolve a circular field with
    more accuracy,

    .. math::

       \phi = \left( x^2 + y^2 \right)^{1/2}

    Build the test case in the following way,

    >>> mesh = Grid2D(dx = 1., dy = 1., nx = 10, ny = 10)
    >>> vel = 1.
    >>> x, y = mesh.cellCenters
    >>> r = numerix.sqrt(x**2 + y**2)
    >>> coeff = CellVariable(mesh = mesh, value = r)
    >>> v, L, b = __AdvectionTerm(1.)._buildMatrix(coeff, SparseMatrix)
    >>> error = CellVariable(mesh=mesh, value=b + 1)
    >>> ans = CellVariable(mesh=mesh, value=b + 1)
    >>> ans[(x > 2) & (x < 8) & (y > 2) & (y < 8)] = 0.123105625618
    >>> print((error <= ans).all())
    True

    The maximum error is large (about 12 %) for the first order advection.

    >>> v, L, b = AdvectionTerm(1.)._buildMatrix(coeff, SparseMatrix)
    >>> error = CellVariable(mesh=mesh, value=b + 1)
    >>> ans = CellVariable(mesh=mesh, value=b + 1)
    >>> ans[(x > 2) & (x < 8) & (y > 2) & (y < 8)] = 0.0201715476598
    >>> print((error <= ans).all())
    True

    The maximum error is 2 % when using a higher order contribution.

    """
    def _getDifferences(self, adjacentValues, cellValues, oldArray, cellToCellIDs, mesh):

        dAP = mesh._cellToCellDistances

##        adjacentGradient = numerix.take(oldArray.grad, cellToCellIDs)
        adjacentGradient = numerix.take(oldArray.grad, mesh._cellToCellIDs, axis=-1)
        adjacentNormalGradient = numerix.dot(adjacentGradient, mesh._cellNormals)
        adjacentUpValues = cellValues + 2 * dAP * adjacentNormalGradient

        cellIDs = numerix.repeat(numerix.arange(mesh.numberOfCells)[numerix.newaxis, ...],
                mesh._maxFacesPerCell, axis=0)
        cellIDs = MA.masked_array(cellIDs, mask = MA.getmask(mesh._cellToCellIDs))
        cellGradient = numerix.take(oldArray.grad, cellIDs, axis=-1)
        cellNormalGradient = numerix.dot(cellGradient, mesh._cellNormals)
        cellUpValues = adjacentValues - 2 * dAP * cellNormalGradient

        cellLaplacian = (cellUpValues + adjacentValues - 2 * cellValues) / dAP**2

        adjacentLaplacian = (adjacentUpValues + cellValues - 2 * adjacentValues) / dAP**2
        adjacentLaplacian = adjacentLaplacian.filled(0)
        cellLaplacian = cellLaplacian.filled(0)

        mm = numerix.where(cellLaplacian * adjacentLaplacian < 0.,
                           0.,
                           numerix.where(abs(cellLaplacian) > abs(adjacentLaplacian),
                                         adjacentLaplacian,
                                         cellLaplacian))

        return FirstOrderAdvectionTerm._getDifferences(self, adjacentValues, cellValues, oldArray, cellToCellIDs, mesh) -  mm * dAP / 2.

class __AdvectionTerm(FirstOrderAdvectionTerm):
    """
    Dummy subclass for tests
    """
    pass

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.terms.abstractConvectionTerm import _AbstractConvectionTerm
from fipy.variables.faceVariable import FaceVariable

__all__ = ["CentralDifferenceConvectionTerm"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class _CentralDifferenceConvectionTermAlpha(FaceVariable):
    def __init__(self, P):
        FaceVariable.__init__(self, P.mesh, value=0.5)

class CentralDifferenceConvectionTerm(_AbstractConvectionTerm):
    r"""

    This :class:`~fipy.terms.term.Term` represents

    .. math::

       \int_V \nabla \cdot (\vec{u} \phi)\,dV \simeq \sum_{f} (\vec{n}
       \cdot \vec{u})_f \phi_f A_f

    where :math:`\phi_f=\alpha_f \phi_P +(1-\alpha_f)\phi_A` and
    :math:`\alpha_f` is calculated using the central differencing scheme.
    For further details see :ref:`sec:NumericalSchemes`.
    """
    def _alpha(self, P):
        return _CentralDifferenceConvectionTermAlpha(P)



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from PyTrilinos import AztecOO

from fipy.solvers.trilinos.preconditioners.preconditioner import Preconditioner

__all__ = ["DomDecompPreconditioner"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class DomDecompPreconditioner(Preconditioner):
    """
    Domain Decomposition preconditioner for Trilinos solvers.

    """

    def _ApplyToSolver(self, solver, matrix):
        solver.SetAztecOption(AztecOO.AZ_precond, AztecOO.AZ_dom_decomp)



from __future__ import unicode_literals
from builtins import object
__docformat__ = 'restructuredtext'

from ...tools import numerix
from ...variables.cellVariable import CellVariable

__all__ = []

class _AbstractTopology(object):
    _concatenatedClass = None

    def __init__(self, mesh):
        self.mesh = mesh

    @property
    def _isOrthogonal(self):
        raise NotImplementedError

    @property
    def _globalNonOverlappingCellIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Does not include the IDs of boundary cells.

        E.g., would return [0, 1, 4, 5] for mesh A

        ```
            A        B
        ------------------
        | 4 | 5 || 6 | 7 |
        ------------------
        | 0 | 1 || 2 | 3 |
        ------------------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.numberOfCells)

    @property
    def _globalOverlappingCellIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Includes the IDs of boundary cells.

        E.g., would return [0, 1, 2, 4, 5, 6] for mesh A

        ```
            A        B
        ------------------
        | 4 | 5 || 6 | 7 |
        ------------------
        | 0 | 1 || 2 | 3 |
        ------------------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.numberOfCells)

    @property
    def _localNonOverlappingCellIDs(self):
        """Return the IDs of the local mesh in isolation.

        Does not include the IDs of boundary cells.

        E.g., would return [0, 1, 2, 3] for mesh A

        ```
            A        B
        ------------------
        | 3 | 4 || 4 | 5 |
        ------------------
        | 0 | 1 || 1 | 2 |
        ------------------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.numberOfCells)

    @property
    def _localOverlappingCellIDs(self):
        """Return the IDs of the local mesh in isolation.

        Includes the IDs of boundary cells.

        E.g., would return [0, 1, 2, 3, 4, 5] for mesh A

        ```
            A        B
        ------------------
        | 3 | 4 || 5 |   |
        ------------------
        | 0 | 1 || 2 |   |
        ------------------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.numberOfCells)

    @property
    def _globalNonOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Does not include the IDs of faces of boundary cells.

        E.g., would return [0, 1, 4, 5, 8, 9, 12, 13, 14, 17, 18, 19]
        for mesh A

        ```
            A   ||   B
        --8---9---10--11--
       17   18  19  20   21
        --4---5----6---7--
       12   13  14  15   16
        --0---1----2---3--
                ||
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.numberOfFaces)

    @property
    def _globalOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Includes the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2, 4, 5, 6, 8, 9, 10, 12, 13,
        14, 15, 17, 18, 19, 20] for mesh A

        ```
            A   ||   B
        --8---9---10--11--
       17   18  19  20   21
        --4---5----6---7--
       12   13  14  15   16
        --0---1----2---3--
                ||
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.numberOfFaces)

    @property
    def _cellProcID(self):
        """
        Return the processor that "owns" each cell.
        Ownership goes to the lowest `procID` of a neighboring cell.
        Boundary faces try not to be duplicated

        E.g., would return [0, 0, 1, 0, 0, 1] for mesh A
                       and [0, 1, 1, 0, 1, 1] for mesh B

        ```
            A        B
        ------------------
        | 4 | 5 || 6 | 7 |
        ------------------
        | 0 | 1 || 2 | 3 |
        ------------------
        ```

        .. note:: Trivial except for parallel meshes
        """
        procID = CellVariable(mesh=self.mesh,
                              value=self.mesh.communicator.procID)
        procID._updateGhosts()

        return procID

    @property
    def _ownedFaceIDs(self):
        """
        Return the local face IDs of the mesh "owned" by the current
        processor.  Ownership goes to the lowest `procID` of a neighboring
        cell.  Boundary faces try not to be duplicated

        E.g., would return [0, 1, 3, 4, 6, 7, 9, 10, 11, 13, 14, 15] for mesh A
                       and [1, 2, 4, 5, 7, 8, 11, 12, 15, 16] for mesh B

        ```
            A   ||   B
        --6---7-----7---8--
       13   14 15/14 15   16
        --3---4-----4---5--
        9   10 11/10 11   12
        --0---1-----1---2--
                ||
        ```

        .. note:: Trivial except for parallel meshes
        """
        minproc = numerix.take(self._cellProcID,
                               self.mesh.faceCellIDs).min(axis=0)

        return numerix.where(minproc == self.mesh.communicator.procID)[0]

    @property
    def _nonOverlappingFaces(self):
        """Return mask of faces on local mesh.

        False for faces only belonging to ghost cells.

        E.g., would return [True, True, False, True, True, False, True,
        True, False, True, True, True, False, True, True, True, False]
        for mesh A

        ```
            A   ||   B
        --6---7----8------
       13   14  15  16   |
        --3---4----5------
        9   10  11  12   |
        --0---1----2------
                ||
        ```
        """
        return (numerix.in1dMA(self.mesh.faceCellIDs[0],
                               self._localNonOverlappingCellIDs).filled(False)
                | numerix.in1dMA(self.mesh.faceCellIDs[1],
                                 self._localNonOverlappingCellIDs).filled(False))

    @property
    def _localNonOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in isolation.

        Does not include the IDs of faces of boundary cells.

        E.g., would return [0, 1, 3, 4, 6, 7, 9, 10, 11, 13, 14, 15]
        for mesh A

        ```
            A   ||   B
        --6---7-----7---8--
       13   14 15/14 15   16
        --3---4-----4---5--
        9   10 11/10 11   12
        --0---1-----1---2--
                ||
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.numberOfFaces)

    @property
    def _localOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in isolation.

        Includes the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11,
        12, 13, 14, 15, 16] for mesh A

        ```
            A   ||   B
        --6---7----8------
       13   14  15  16   |
        --3---4----5------
        9   10  11  12   |
        --0---1----2------
                ||
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.numberOfFaces)

    @property
    def _vertexCellIDs(self):
        """Return cell IDs bounded by each vertex

        E.g., would return

        [[ 0  0  1  0  0  1  2  2  3]
         [--  1 --  2  1  3 --  3 --]
         [-- -- -- --  2 -- -- -- --]
         [-- -- -- --  3 -- -- -- --]]

        given

        ```
        6-------7-------8
        |       |       |
        |   2   |   3   |
        |       |       |
        3-------4-------5
        |       |       |
        |   0   |   1   |
        |       |       |
        0-------1-------2
        ```
        """
        return numerix.invert_indices(self.mesh._cellVertexIDs)

    @property
    def _vertexFaceIDs(self):
        """Return face IDs bounded by each vertex

        E.g., would return

        [[  0   0   1   2   2   3   4   4   5]
         [  6   1   8   6   3   8   9   5  11]
         [ --   7  --   9   7  11  --  10  --]
         [ --  --  --  --  10  --  --  --  --]]

        given

        ```
        6---4---7---5---8
        |       |       |
        9       10      11
        |       |       |
        3---2---4---3---5
        |       |       |
        6       7       8
        |       |       |
        0---0---1---1---2
        ```
        """
        return numerix.invert_indices(self.mesh.faceVertexIDs)

    # abstract element types mutually understood by FiPy and other meshing systems
    # (VTK, Gmsh, etc.)
    _elementTopology = dict([(k, v) for (v, k) in enumerate(("vertex",
                                                             "line",
                                                             "triangle",
                                                             "quadrangle",
                                                             "pixel",
                                                             "polygon",
                                                             "tetrahedron",
                                                             "hexahedron",
                                                             "voxel",
                                                             "prism",
                                                             "pyramid",
                                                             "unknown"))])

    @property
    def _cellTopology(self):
        """return a map of the topology of each cell of grid"""
        raise NotImplementedError



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.tools import numerix

from fipy.boundaryConditions.boundaryCondition import BoundaryCondition
from fipy.boundaryConditions.fixedFlux import FixedFlux
from fipy.boundaryConditions.fixedValue import FixedValue

__all__ = ["NthOrderBoundaryCondition"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class NthOrderBoundaryCondition(BoundaryCondition):
    """

    This boundary condition is generally used in conjunction with a
    `ImplicitDiffusionTerm` that has multiple coefficients.  It does not
    have any direct effect on the solution matrices, but its derivatives
    do.

    """

    def __init__(self, faces, value, order):
        """
        Creates an `NthOrderBoundaryCondition`.

        Parameters
        ----------
        faces : :obj:`~fipy.variables.faceVariable.FaceVariable` of :obj:`bool`
            Mask of faces where this condition applies.
        value : float
            Value to impose.
        order : int
            Order of the boundary condition. An `order` of `0`
            corresponds to a `FixedValue` and an `order` of `1` corresponds to
            a `FixedFlux`. Even and odd orders behave like `FixedValue` and `FixedFlux` objects,
            respectively, but apply to higher order terms.
        """
        self.order = order
        self.derivative = {}
        BoundaryCondition.__init__(self, faces, value)

    def _buildMatrix(self, SparseMatrix, Ncells, MaxFaces, coeff):
        """Leave **L** and **b** unchanged

        Parameters
        ----------
        SparseMatrix
            *unused*
        Ncells
            *unused*
        MaxFaces
            *unused*
        coeff
            *unused*
        """
        return (0, 0)

    def _getDerivative(self, order):
        newOrder = self.order - order
        if newOrder not in self.derivative:
            if newOrder > 1:
                self.derivative[newOrder] = NthOrderBoundaryCondition(self.faces, self.value, newOrder)
            elif newOrder == 1:
                self.derivative[newOrder] = FixedFlux(self.faces, self.value)
            elif newOrder == 0:
                self.derivative[newOrder] = FixedValue(self.faces, self.value)
            else:
                self.derivative[newOrder] = None

        return self.derivative[newOrder]



from __future__ import unicode_literals
from fipy.solvers.scipy.scipySolver import _ScipySolver
from pyamg import solve
import os
from fipy.tools import numerix

__all__ = ["LinearGeneralSolver"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class LinearGeneralSolver(_ScipySolver):
    """
    The `LinearGeneralSolver` is an interface to the generic PyAMG,
    which solves the arbitrary system Ax=b with the best out-of-the box
    choice for a solver. See `pyAMG.solve` for details.
    """

    def _solve_(self, L, x, b):
        if 'FIPY_VERBOSE_SOLVER' in os.environ:
            verbosity = True
        else:
            verbosity = False

        return solve(L.matrix, b, verb=verbosity, tol=self.tolerance)



from __future__ import division
from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.variables.faceGradVariable import _FaceGradVariable
from fipy.tools import inline
from fipy.tools import numerix

class _ModFaceGradVariable(_FaceGradVariable):
    def __init__(self, var, modIn):
        _FaceGradVariable.__init__(self, var)
        self.modIn = modIn

    if inline.doInline:
        def _calcValue(self):

            id1, id2 = self.mesh._adjacentCellIDs

            tangents1 = self.mesh._faceTangents1
            tangents2 = self.mesh._faceTangents2

            val = self._array.copy()

            inline._runIterateElementInline(self.modIn + """
            int j;
            double t1grad1, t1grad2, t2grad1, t2grad2, N;
            int ID1 = ITEM(id1, i, NULL);
            int ID2 = ITEM(id2, i, NULL);
            N = mod(ITEM(var, ID2, NULL) - ITEM(var, ID1, NULL)) / ITEM(dAP, i, NULL);

            t1grad1 = t1grad2 = t2grad1 = t2grad2 = 0.;

            t1grad1 += ITEM(tangents1, i, vec) * ITEM(cellGrad, ID1, vec);
            t1grad2 += ITEM(tangents1, i, vec) * ITEM(cellGrad, ID2, vec);
            t2grad1 += ITEM(tangents2, i, vec) * ITEM(cellGrad, ID1, vec);
            t2grad2 += ITEM(tangents2, i, vec) * ITEM(cellGrad, ID2, vec);

            ITEM(val, i, vec) =  ITEM(normals, i, vec) * N;
            ITEM(val, i, vec) += ITEM(tangents1, i, vec) * (t1grad1 + t1grad2) / 2.;
            ITEM(val, i, vec) += ITEM(tangents2, i, vec) * (t2grad1 + t2grad2) / 2.;

            """, tangents1 = tangents1,
                tangents2 = tangents2,
                cellGrad = self.var.grad.numericValue,
                normals = numerix.array(self.mesh._orientedFaceNormals),
                id1 = numerix.array(id1),
                id2 = numerix.array(id2),
                dAP = numerix.array(self.mesh._cellDistances),
                var = self.var.numericValue,
                val = val,
                ni = tangents1.shape[1],
                shape=numerix.array(numerix.shape(val)))

            return self._makeValue(value = val)
    else:
        def _calcValue(self):
            dAP = self.mesh._cellDistances
            id1, id2 = self.mesh._adjacentCellIDs
            N = (numerix.take(self.var, id2) - numerix.take(self.var, id1)) / dAP
            normals = self.mesh._orientedFaceNormals

            tangents1 = self.mesh._faceTangents1
            tangents2 = self.mesh._faceTangents2
            cellGrad = self.var.grad.numericValue

            grad1 = numerix.take(cellGrad, id1, axis=1)
            grad2 = numerix.take(cellGrad, id2, axis=1)
            t1grad1 = numerix.sum(tangents1*grad1, 0)
            t1grad2 = numerix.sum(tangents1*grad2, 0)
            t2grad1 = numerix.sum(tangents2*grad1, 0)
            t2grad2 = numerix.sum(tangents2*grad2, 0)

            T1 = (t1grad1 + t1grad2) / 2.
            T2 = (t2grad1 + t2grad2) / 2.

            return normals * N + tangents1 * T1 + tangents2 * T2



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.terms.diffusionTerm import DiffusionTerm as ImplicitDiffusionTerm

__all__ = ["ImplicitDiffusionTerm"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]



from __future__ import unicode_literals
from builtins import object
__docformat__ = 'restructuredtext'

__all__ = ["AbstractMatplotlibViewer"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

from future.builtins import super

from fipy.viewers.viewer import AbstractViewer

def _isnotebook():
    """return True if running in a jupyter notebook
     https://stackoverflow.com/a/39662359/2019542
    """
    try:
        import IPython

        shell = IPython.get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except (ImportError, NameError):
        return False      # Probably standard Python interpreter

def _isretina():
    """return `True` if jupyter notebook configuration
    `InlineBackend.figure_formats` contains `retina` or
    `InlineBackend.figure_format` is set to `retina`
    """
    isretina = False

    try:
        import IPython

        cfg = IPython.get_ipython().config
        if "InlineBackend" in cfg:
            inbk = cfg["InlineBackend"]
            if "figure_formats" in inbk:
                if "retina" in inbk["figure_formats"]:
                    isretina = True
            if not isretina and ("figure_format" in inbk):
                isretina = inbk["figure_format"] == "retina"
    except ImportError:
        pass
    return isretina

class AbstractMatplotlibViewer(AbstractViewer):
    """
    .. attention:: This class is abstract. Always create one of its subclasses.

    The `AbstractMatplotlibViewer` is the base class for the viewers that use the
    Matplotlib_ python plotting package.

    .. _Matplotlib: http://matplotlib.sourceforge.net/
    """

    def __init__(self, vars, title=None, figaspect=1.0, cmap=None, colorbar=None, axes=None, log=False, **kwlimits):
        """
        Create a `AbstractMatplotlibViewer`.

        Parameters
        ----------
        vars : ~fipy.variables.cellVariable.CellVariable or list
            the `Variable` objects to display.
        title : str, optional
            displayed at the top of the `Viewer` window
        figaspect : float, optional
            desired aspect ratio of figure. If arg is a number, use that aspect
            ratio. If arg is `auto`, the aspect ratio will be determined from
            the Variable's mesh.
        xmin, xmax, ymin, ymax, datamin, datamax : float, optional
            displayed range of data. A 1D `Viewer` will only use *xmin* and
            *xmax*, a 2D viewer will also use *ymin* and *ymax*. All
            viewers will use *datamin* and *datamax*. Any limit set to a
            (default) value of `None` will autoscale.
        cmap : ~matplotlib.colors.Colormap, optional
            the :class:`~matplotlib.colors.Colormap`.
            Defaults to `matplotlib.cm.jet`
        colorbar : bool, optional
            plot a color bar in specified orientation if not `None`
        axes : ~matplotlib.axes.Axes, optional
            if not `None`, `vars` will be plotted into this
            :ref:`Matplotlib` :class:`~matplotlib.axes.Axes` object
        log : bool, optional
          whether to logarithmically scale the data
        """
        if self.__class__ is AbstractMatplotlibViewer:
            raise NotImplementedError("can't instantiate abstract base class")

        super(AbstractMatplotlibViewer, self).__init__(vars=vars, title=None, **kwlimits)

        from matplotlib import pyplot as plt

        plt.ion()

        if axes is None:
            w, h = plt.figaspect(self.figaspect(figaspect, colorbar))
            self._fig = plt.figure(figsize=(w, h))
            self._axes = plt.gca()
        else:
            self._axes = axes
            self._fig = axes.get_figure()

        self._mappable_ = None

        import matplotlib
        # Set the colormap and norm to correspond to the data for which
        # the colorbar will be used.
        if cmap is None:
            self._cmap = matplotlib.cm.jet
        else:
            self._cmap = cmap

        self._norm = None
        self.log = log

        if colorbar:
            self._colorbar = self.fig.colorbar(mappable=self._mappable,
                                               orientation=colorbar,
                                               label=self.vars[0].name)

        self.title = title

    @property
    def axes(self):
        """The :ref:`Matplotlib` :class:`~matplotlib.axes.Axes`."""
        return getattr(self, "_axes", None)

    @property
    def cmap(self):
        """The :ref:`Matplotlib` :class:`~~matplotlib.colors.Colormap`."""
        return self._cmap

    @cmap.setter
    def cmap(self, value):
        self._cmap = value
        self._mappable.set_cmap(value)

    @property
    def colorbar(self):
        """The :ref:`Matplotlib` :class:`~matplotlib.colorbar.Colorbar`."""
        return getattr(self, "_colorbar", None)

    @property
    def fig(self):
        """The :ref:`Matplotlib` :class:`~matplotlib.figure.Figure`."""
        return self._fig

    @property
    def id(self):
        """The :ref:`Matplotlib` :class:`~matplotlib.figure.Figure` number."""
        return self.fig.number

    @property
    def log(self):
        """Whether data has logarithmic scaling (:class:`bool`).
        """
        from matplotlib import colors
        return isinstance(self._norm, colors.LogNorm)

    @log.setter
    def log(self, value):
        zmin, zmax = self._autoscale(vars=self.vars,
                                     datamin=self._getLimit(('datamin', 'zmin')),
                                     datamax=self._getLimit(('datamax', 'zmax')))

        from matplotlib import colors
        if value:
            self._norm = colors.LogNorm(vmin=zmin, vmax=zmax)
        else:
            self._norm = colors.Normalize(vmin=zmin, vmax=zmax)

    def _make_mappable(self):
        import matplotlib
        mappable = matplotlib.cm.ScalarMappable(norm=self._norm, cmap=self.cmap)
        # ignored, but needed for matplotlib < 3.0 (Py2k)
        mappable.set_array(self.vars[0].value)

        return mappable

    @property
    def _mappable(self):
        """The :class:`~matplotlib.cm.ScalarMappable` between data and rendering.
        """
        if self._mappable_ is None:
            self._mappable_ = self._make_mappable()
        return self._mappable_

    @property
    def _norm(self):
        """The :ref:`Matplotlib` :class:`~matplotlib.colors.Normalize`."""
        return self._norm_

    @_norm.setter
    def _norm(self, value):
        self._norm_ = value
        self._mappable.set_norm(value)
        if self.colorbar is not None:
            self.colorbar.update_normal(self._mappable)

    @AbstractViewer.title.setter
    def title(self, value):
        # attrociousness needed to call inherited setter
        # https://stackoverflow.com/q/3336767/2019542
        # https://gist.github.com/Susensio/979259559e2bebcd0273f1a95d7c1e79
        super(AbstractMatplotlibViewer, type(self)).title.fset(self, value)
        if self.axes is not None:
            self.axes.set_title(self.title)

    def figaspect(self, figaspect, colorbar):
        return figaspect

    def plot(self, filename = None):
        from matplotlib import pyplot as plt

        plt.ioff()

        self._plot()

        plt.draw()

        try:
            self.fig.canvas.flush_events()
        except NotImplementedError:
            pass

        plt.ion()

        if _isnotebook():
            # plots don't animate in the notebook unless we
            # explicitly clear_output and display
            from IPython.display import display, clear_output

            clear_output(wait=True)
            display(self)
        else:
            self.fig.show()

        if filename is not None:
            self.fig.savefig(filename)

    def _validFileExtensions(self):
        return ["""
        Matplotlib has no reliable way to determine
        valid file extensions. Either guess, or see
        <http://matplotlib.sourceforge.net/faq/installing_faq.html#backends>
        and then guess. Yes, this is lame.
        """]

#         filetypes = plt.figure(self.id).canvas.filetypes
#         return [".%s" % key for key in filetypes.keys()]

    def _repr_png_(self):
        """Render as a PNG for IPython notebook, per `display_protocol.ipynb`

        Invoke with `display(myViewer)`
        """
        from IPython.core.pylabtools import print_figure, retina_figure

        if _isretina():
            return retina_figure(self.fig)
        else:
            return print_figure(self.fig, "png")

    @classmethod
    def _doctest_extra(cls):
        return ("""
            >>> viewer.cmap = "ocean"
            >>> viewer.log = True
        """ + super()._doctest_extra())

if __name__ == "__main__":
    import fipy.tests.doctestPlus
    fipy.tests.doctestPlus.execButNoTest()



from __future__ import unicode_literals
from fipy.tools import numerix
from fipy.tools.comms.abstractCommWrapper import AbstractCommWrapper

__all__ = ["DummyComm"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class DummyComm(AbstractCommWrapper):
    @property
    def procID(self):
        return 0
        
    @property
    def Nproc(self):
        return 1



from __future__ import print_function
from distutils.core import Command
import os

__all__ = ["upload_products"]

class upload_products(Command):
    description = "upload FiPy compressed archives to website(s)"
    
    user_options = [('pdf', None, "upload the PDF variant of the documentation"),
                    ('html', None, "upload the HTML variant of the documentation"),
                    ('tarball', None, "upload the .tar.gz source distribution"),
                    ('winzip', None, "upload the .win32.zip distribution"),
                   ]

    def initialize_options (self):
        self.pdf = 0
        self.html = 0
        self.tarball = 0
        self.winzip = 0

    def finalize_options (self):
        pass

    def run(self):
        if self.pdf:
            fname = 'dist/fipy-%s.pdf' % self.distribution.metadata.get_version()
            print("setting permissions of manual...")
            os.system('chmod -R g+w documentation/_build/latex/fipy.pdf')
            
            print("linking manual to `dist/`...")
            os.system('mkdir dist/')
            os.system('ln -f documentation/_build/latex/fipy.pdf %s' % fname)
            
            print("uploading pdf...")
            os.system('rsync -pgoDLC -e ssh %s %s/download/' % (fname, os.environ['FIPY_WWWHOST']))

        if self.html:
            print("setting group and ownership of web pages...")
            os.system('chmod -R g+w documentation/_build/html/')
            
            print("uploading web pages...")
            # The -t flag (implicit in -a) is suddenly causing problems
            # os.system('rsync -aLC -e ssh %s %s'%('documentation/www/', os.environ['FIPY_WWWHOST']))
            os.system('rsync -rlpgoDLC -e ssh %s %s' % ('documentation/_build/html/', os.environ['FIPY_WWWHOST']))

        if self.tarball:
            fname = 'dist/FiPy-%s.tar.gz' % self.distribution.metadata.get_version()
            print("setting permissions for %s ..." % fname)
            os.system('chmod -R g+w %s' % fname)

            print("uploading tarball...")
            os.system('rsync -pgoDLC -e ssh %s %s/download/' % (fname, os.environ['FIPY_WWWHOST']))

        if self.winzip:
            fname = 'dist/FiPy-%s.win32.zip' % self.distribution.metadata.get_version()
            print("setting permissions for %s ..." % fname)
            os.system('chmod -R g+w %s' % fname)
            
            print("uploading winzip...")
            os.system('rsync -pgoDLC -e ssh %s %s/download/' % (fname, os.environ['FIPY_WWWHOST']))

        print("activating web pages...")
        os.system(os.environ['FIPY_WWWACTIVATE'])



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.tools.numerix import random
from fipy.variables.noiseVariable import NoiseVariable

__all__ = ["UniformNoiseVariable"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class UniformNoiseVariable(NoiseVariable):
    r"""
    Represents a uniform distribution of random numbers.

    We generate noise on a uniform Cartesian mesh

    >>> from fipy.meshes import Grid2D
    >>> noise = UniformNoiseVariable(mesh=Grid2D(nx=100, ny=100))

    and histogram the noise

    >>> from fipy.variables.histogramVariable import HistogramVariable
    >>> histogram = HistogramVariable(distribution=noise, dx=0.01, nx=120, offset=-.1)

    >>> if __name__ == '__main__':
    ...     from fipy import Viewer
    ...     viewer = Viewer(vars=noise,
    ...                     datamin=0, datamax=1)
    ...     histoplot = Viewer(vars=histogram)

    >>> from builtins import range
    >>> for i in range(10):
    ...     noise.scramble()
    ...     if __name__ == '__main__':
    ...         viewer.plot()
    ...         histoplot.plot()

    .. image:: fipy/variables/uniform.*
       :scale: 25
       :align: center
       :alt: random values with a uniform distribution

    .. image:: fipy/variables/uni-histogram.*
       :scale: 25
       :align: center
       :alt: histogram of random values with a uniform distribution
    """
    def __init__(self, mesh, name = '', minimum = 0., maximum = 1., hasOld = 0):
        """
        Parameters
        ----------
        mesh : ~fipy.meshes.mesh.Mesh
            The mesh on which to define the noise.
        minimum : float
            The minimum (not-inclusive) value of the distribution.
        maximum : float
            The maximum (not-inclusive) value of the distribution.
        """
        self.minimum = minimum
        self.maximum = maximum
        NoiseVariable.__init__(self, mesh = mesh, name = name, hasOld = hasOld)

    def random(self):
        return random.uniform(self.minimum, self.maximum,
                              size=[self.mesh.globalNumberOfCells])

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



from builtins import object
__docformat__ = 'restructuredtext'

from fipy.tools import numerix

__all__ = ["AbstractCommWrapper"]

class AbstractCommWrapper(object):
    """MPI Communicator wrapper
    
    Encapsulates capabilities needed for possibly parallel operations. 
    Some capabilities are not parallel.
    
    """
    
    def __init__(self):
        pass
        
    def __repr__(self):
        return "%s()" % self.__class__.__name__

    @property
    def procID(self):
        raise NotImplementedError
        
    @property
    def Nproc(self):
        raise NotImplementedError
        
    def Barrier(self):
        pass

    def all(self, a, axis=None):
        return a.all(axis=axis)

    def any(self, a, axis=None):
        return a.any(axis=axis)

    def allclose(self, a, b, rtol=1.e-5, atol=1.e-8):
        return numerix.allclose(a, b, rtol=rtol, atol=atol)

    def allequal(self, a, b):
        return numerix.allequal(a, b)

    def bcast(self, obj, root=0):
        return obj

    def allgather(self, obj):
        return obj

    def sum(self, a, axis=None):
        return a.sum(axis=axis)

    def __getstate__(self):
        return {'dummy': 0}

    def __setstate__(self, dict):
        self.__init__()
        
    def Norm2(self, vec):
        return numerix.L2norm(vec)

    @staticmethod
    def _tolist(vec):
        try:
            return list(vec)
        except TypeError:
            return [vec]

    def MaxAll(self, vec):
        return max(self._tolist(vec))
        
    def MinAll(self, vec):
        return min(self._tolist(vec))



"""
The iterative solvers may output warnings if the solution is considered
unsatisfactory. If you are not interested in these warnings, you can invoke
python with a warning filter such as::

    $ python -Wignore::fipy.SolverConvergenceWarning myscript.py

If you are extremely concerned about your preconditioner for some reason, you
can abort whenever it has problems with::

    $ python -Werror::fipy.PreconditionerWarning myscript.py

"""
from __future__ import division
from __future__ import unicode_literals
from builtins import object
from builtins import str
__docformat__ = 'restructuredtext'

from fipy.tools import numerix

__all__ = ["SolverConvergenceWarning", "MaximumIterationWarning",
           "PreconditionerWarning", "IllConditionedPreconditionerWarning",
           "PreconditionerNotPositiveDefiniteWarning", "MatrixIllConditionedWarning",
           "StagnatedSolverWarning", "ScalarQuantityOutOfRangeWarning", "Solver"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class SolverConvergenceWarning(Warning):
    def __init__(self, solver, iter, relres):
        self.solver = solver
        self.iter = iter
        self.relres = relres

    def __str__(self):
        return "%s failed. Iterations: %g. Relative error: %g" % (str(self.solver), self.iter, self.relres)

class MaximumIterationWarning(SolverConvergenceWarning):
    def __str__(self):
        return "Iterations: %g. Relative error: %g" % (self.iter, self.relres)

class PreconditionerWarning(SolverConvergenceWarning):
    pass

class IllConditionedPreconditionerWarning(PreconditionerWarning):
    def __str__(self):
        return "The system involving the preconditioner was ill-conditioned. Relative error: %g" % (self.relres)

class PreconditionerNotPositiveDefiniteWarning(PreconditionerWarning):
    def __str__(self):
        return "The preconditioning matrix does not appear to be positive definite. Relative error: %g" % (self.relres)

class MatrixIllConditionedWarning(SolverConvergenceWarning):
    def __str__(self):
        return "The matrix appears to be very ill-conditioned. Relative error: %g" % (self.relres)

class StagnatedSolverWarning(SolverConvergenceWarning):
    def __str__(self):
        return "The solver stagnated. Iterations: %g. Relative error: %g" % (self.iter, self.relres)

class ScalarQuantityOutOfRangeWarning(SolverConvergenceWarning):
    def __str__(self):
        return "A scalar quantity became too small or too large to continue computing. Iterations: %g. Relative error: %g" % (self.iter, self.relres)

class Solver(object):
    """
    The base `LinearXSolver` class.

    .. attention:: This class is abstract. Always create one of its subclasses.
    """

    def __init__(self, tolerance=1e-10, iterations=1000, precon=None):
        """
        Create a `Solver` object.

        Parameters
        ----------
        tolerance : float
            Required error tolerance.
        iterations : int
            Maximum number of iterative steps to perform.
        precon
            Preconditioner to use.  Not all solver suites support
            preconditioners.
        """
        if self.__class__ is Solver:
            raise NotImplementedError("can't instantiate abstract base class")

        self.tolerance = tolerance
        self.iterations = iterations

        self.preconditioner = precon

    def _storeMatrix(self, var, matrix, RHSvector):
        self.var = var
        self.matrix = matrix
        self.RHSvector = RHSvector

    def _solve(self):
        raise NotImplementedError

    def _solve_(self, L, x, b):
        raise NotImplementedError

    def _applyUnderRelaxation(self, underRelaxation=None):
        if underRelaxation is not None:
            self.matrix.putDiagonal(numerix.asarray(self.matrix.takeDiagonal()) / underRelaxation)
            self.RHSvector += (1 - underRelaxation) * self.matrix.takeDiagonal() * numerix.array(self.var).flatten()

    def _calcResidualVector(self, residualFn=None):
        if residualFn is not None:
            return residualFn(self.var, self.matrix, self.RHSvector)
        else:
            Lx = self.matrix * numerix.array(self.var).flatten()

            return Lx - self.RHSvector

    def _calcResidual(self, residualFn=None):
        if residualFn is not None:
            return residualFn(self.var, self.matrix, self.RHSvector)
        else:
            return numerix.L2norm(self._calcResidualVector())

    def _calcRHSNorm(self):
        return numerix.L2norm(self.RHSvector)

    _warningList = (ScalarQuantityOutOfRangeWarning,
                    StagnatedSolverWarning,
                    MatrixIllConditionedWarning,
                    PreconditionerNotPositiveDefiniteWarning,
                    IllConditionedPreconditionerWarning,
                    MaximumIterationWarning)

    def _raiseWarning(self, info, iter, relres):
        # info is negative, so we list in reverse order so that
        # info can be used as an index from the end

        if info < 0:
            # is stacklevel=5 always what's needed to get to the user's scope?
            import warnings
            warnings.warn(self._warningList[info](self, iter, relres), stacklevel=5)

    def __repr__(self):
        return '%s(tolerance=%g, iterations=%g)' \
            % (self.__class__.__name__, self.tolerance, self.iterations)

    def _canSolveAsymmetric(self):
        return True

    def __enter__(self):
        # The __enter__() and __exit__() methods allow
        # solver objects to be used within a context manager.
        # This enables one to write:
        #
        #     with ExampleSolver() as solver:
        #         eq.solve(var=phi, solver=solver)
        #
        # For subclasses that do not define __enter__()
        # and __exit__() methods, the above is equivalent to:
        #
        #     eq.solve(var=phi, solver=ExampleSolver())

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass



from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
__docformat__ = 'restructuredtext'

import os
import subprocess
import sys
import tempfile
import time

from fipy.viewers.viewer import AbstractViewer

__all__ = ["MayaviClient"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class MayaviClient(AbstractViewer):
    """
    The `MayaviClient` uses the Mayavi_ python plotting package.

    .. _Mayavi: http://code.enthought.com/projects/mayavi

    """

    def __init__(self, vars, title=None, daemon_file=None, fps=1.0, **kwlimits):
        """Create a `MayaviClient`.

        Parameters
        ----------
        vars : ~fipy.variables.cellVariable.CellVariable or list
            `CellVariable` objects to plot
        title : str, optional
            displayed at the top of the `Viewer` window
        xmin, xmax, ymin, ymax, zmin, zmax, datamin, datamax : float, optional
            displayed range of data. A 1D `Viewer` will only use `xmin` and
            `xmax`, a 2D viewer will also use `ymin` and `ymax`, and so on. All
            viewers will use `datamin` and `datamax`. Any limit set to a
            (default) value of `None` will autoscale.
        daemon_file : str, optional
            the path to the script to run the separate Mayavi viewer process.
            Defaults to `fipy/viewers/mayaviViewer/mayaviDaemon.py`
        fps : float, optional
            frames per second to attempt to display
        """
        self._fps = fps

        self._vtkdir = tempfile.mkdtemp()
        self._vtkcellfname = os.path.join(self._vtkdir, "cell.vtk")
        self._vtkfacefname = os.path.join(self._vtkdir, "face.vtk")
        self._vtklockfname = os.path.join(self._vtkdir, "lock")

        from fipy.viewers.vtkViewer import VTKCellViewer, VTKFaceViewer

        try:
            self._vtkCellViewer = VTKCellViewer(vars=vars)
            cell_vars = self._vtkCellViewer.vars
        except TypeError:
            self._vtkCellViewer = None
            cell_vars = []

        try:
            self._vtkFaceViewer = VTKFaceViewer(vars=vars)
            face_vars = self._vtkFaceViewer.vars
        except TypeError:
            self._vtkFaceViewer = None
            face_vars = []

        AbstractViewer.__init__(self, vars=cell_vars + face_vars, title=title, **kwlimits)

        self.plot()

        from pkg_resources import Requirement, resource_filename
        daemon_file = (daemon_file
                       or resource_filename(Requirement.parse("FiPy"),
                                            "fipy/viewers/mayaviViewer/mayaviDaemon.py"))

        pyth = sys.executable or "python"

        cmd = [pyth,
               daemon_file,
               "--lock",
               self._vtklockfname,
               "--fps",
               str(self._fps)]

        if self._vtkCellViewer is not None:
            cmd += ["--cell", self._vtkcellfname]

        if self._vtkFaceViewer is not None:
            cmd += ["--face", self._vtkfacefname]


        cmd += self._getLimit('xmin')
        cmd += self._getLimit('xmax')
        cmd += self._getLimit('ymin')
        cmd += self._getLimit('ymax')
        cmd += self._getLimit('zmin')
        cmd += self._getLimit('zmax')
        cmd += self._getLimit('datamin')
        cmd += self._getLimit('datamax')

        self._daemon = subprocess.Popen(cmd)

    @property
    def fps(self):
        """The frames per second to attempt to display."""
        return self._fps

    def __del__(self):
        for fname in [self._vtkcellfname, self._vtkfacefname, self._vtklockfname]:
            if fname and os.path.isfile(fname):
                os.unlink(fname)
        os.rmdir(self._vtkdir)

    def _getLimit(self, key, default=None):
        """
        Return the limit associated with the key

        .. Note::

           `MayaviClient` does not need the generality of multiple keys
           because it is always 3D

        Parameters
        ----------
        key : str
            dictionary key that identifies the limit of interest

        Returns
        -------
        float or None
            the value of the limit
        """
        lim = AbstractViewer._getLimit(self, key, default=None)
        if lim is not None:
            return ["--%s" % key, str(lim)]
        else:
            return []

    def plot(self, filename=None):
        start = time.time()
        plotted = False
        while not plotted:
            if not os.path.isfile(self._vtklockfname):
                if self._vtkCellViewer is not None:
                    self._vtkCellViewer.plot(filename=self._vtkcellfname)
                if self._vtkFaceViewer is not None:
                    self._vtkFaceViewer.plot(filename=self._vtkfacefname)
                with open(self._vtklockfname, 'w') as lock:
                    if filename is not None:
                        lock.write(filename)
                plotted = True

            if (time.time() - start > 30. / self._fps) and not plotted:
                print("viewer: NOT READY")
                start = time.time()
        if not plotted:
            print("viewer: SKIPPED")

    def _validFileExtensions(self):
        return [".png", ".jpg", ".bmp", ".tiff", ".ps", ".eps", ".pdf", ".rib", ".oogl", ".iv", ".vrml", ".obj"]

    @classmethod
    def _doctest_body(cls):
        return (cls._test1D()
                + cls._test2D()
                + cls._test2Dirregular()
                + cls._test3D())

    @classmethod
    def _doctest_extra(cls):
        return ""

if __name__ == "__main__":
    import fipy.tests.doctestPlus
    fipy.tests.doctestPlus.execButNoTest()



from fipy.solvers.petsc.linearLUSolver import *
from fipy.solvers.petsc.linearPCGSolver import *
from fipy.solvers.petsc.linearGMRESSolver import *
from fipy.solvers.petsc.linearBicgSolver import *
from fipy.solvers.petsc.linearCGSSolver import *
from fipy.solvers.petsc.dummySolver import *

DefaultSolver = LinearGMRESSolver
DefaultAsymmetricSolver = LinearGMRESSolver
DummySolver = DummySolver
GeneralSolver = DefaultSolver

__all__ = ["DefaultSolver",
           "DummySolver",
           "DefaultAsymmetricSolver",
           "GeneralSolver"]
           
__all__.extend(linearLUSolver.__all__)
__all__.extend(linearPCGSolver.__all__)
__all__.extend(linearGMRESSolver.__all__)
__all__.extend(linearBicgSolver.__all__)
__all__.extend(linearCGSSolver.__all__)



from __future__ import division
from __future__ import unicode_literals
from fipy.steppers.stepper import Stepper

__all__ = ["PIDStepper"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class PIDStepper(Stepper):
    """
    Adaptive stepper using a PID controller, based on::

        @article{PIDpaper,
           author =  {A. M. P. Valli and G. F. Carey and A. L. G. A. Coutinho},
           title =   {Control strategies for timestep selection in finite element
                      simulation of incompressible flows and coupled
                      reaction-convection-diffusion processes},
           journal = {Int. J. Numer. Meth. Fluids},
           volume =  47,
           year =    2005,
           pages =   {201-231},
        }
    """
    def __init__(self, vardata=(), proportional=0.075, integral=0.175, derivative=0.01):
        Stepper.__init__(self, vardata=vardata)

        self.proportional = proportional
        self.integral = integral
        self.derivative = derivative

        self.error = [1., 1., 1.]
        self.nrej = 0

    def _step(self, dt, dtPrev, sweepFn, failFn, *args, **kwargs):
        while True:
            self.error[2] = sweepFn(vardata=self.vardata, dt=dt, *args, **kwargs)

            # omitting nsa > nsaMax check since it's unclear from
            # the paper what it's supposed to do
            if self.error[2] > 1. and dt > self.dtMin:
                # reject the timestep
                failFn(vardata=self.vardata, dt=dt, *args, **kwargs)

                self.nrej += 1

                for var, eqn, bcs in self.vardata:
                    var.setValue(var.old)

                factor = min(1. / self.error[2], 0.8)

                dt = self._lowerBound(factor * dt)

                dtPrev = dt**2 / dtPrev
            else:
                # step succeeded
                break

        dtNext = dtPrev * ((self.error[1] / self.error[2])**self.proportional
                           * (1. / self.error[2])**self.integral
                           * (self.error[1]**2 / (self.error[2] * self.error[0]))**self.derivative)

        self.error[0] = self.error[1]
        self.error[1] = self.error[2]

        return dt, dtNext



from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
from setuptools.command.test import test as _test
from future.utils import text_to_native_str
from future.utils import string_types
import unittest
import warnings
import sys

__all__ = [text_to_native_str("test")]

def _nativize_all(t):
    def _nativize(s):
        if isinstance(s, string_types):
            s = text_to_native_str(s)
        return s

    return tuple([_nativize(s) for s in t])

class DeprecationErroringTestProgram(unittest.TestProgram):
    """`TestProgram` that overrides inability of standard
    `TestProgram` to throw errors on `DeprecationWarning`
    """
    def runTests(self):
        self.warnings = None

        with warnings.catch_warnings():
            warnings.simplefilter(action="error", category=DeprecationWarning)

            # Don't raise noisy errors in
            # tvtk/tvtk_classes.zip/tvtk_classes/abstract_array.py or
            # tvtk/tvtk_classes.zip/tvtk_classes/algorithm.py
            warnings.filterwarnings(action="default", category=DeprecationWarning,
                                    message="invalid escape sequence.*")

            # Don't raise errors in skfmm.pfmm
            # due to deprecation of np.int in NumPy 1.20
            warnings.filterwarnings(action="default", category=DeprecationWarning,
                                    message="`np\.int` is a deprecated alias for the builtin `int`.*",
                                    module="skfmm\.pfmm")

            super(DeprecationErroringTestProgram, self).runTests()

class test(_test):
    description = str(_test.description) + ", for FiPy and its examples"

    # List of option tuples: long name, short name (None if no short
    # name), and help string.
    user_options = _test.user_options + [
        ('inline', None, "run FiPy with inline compilation enabled"),
        ('pythoncompiled=', None, "directory in which to put weave's work product"),
        ('Trilinos', None, "run FiPy using Trilinos solvers"),
        ('Pysparse', None, "run FiPy using Pysparse solvers (default)"),
        ('trilinos', None, "run FiPy using Trilinos solvers"),
        ('pysparse', None, "run FiPy using Pysparse solvers (default)"),
        ('scipy', None, "run FiPy using SciPy solvers"),
        ('Scipy', None, "run FiPy using SciPy solvers"),
        ('petsc', None, "run FiPy using PETSc solvers"),
        ('no-pysparse', None, "run FiPy without using the Pysparse solvers"),
        ('pyamg', None, "run FiPy without using the PyAMG solvers"),
        ('pyamgx', None, "run FiPy using the pyamgx solvers"),
        ('all', None, "run all non-interactive FiPy tests (default)"),
        ('really-all', None, "run *all* FiPy tests (including those requiring user input)"),
        ('examples', None, "test FiPy examples"),
        ('modules', None, "test FiPy code modules"),
        ('viewers', None, "test FiPy viewer modules (requires user input)"),
        ('cache', None, "run FiPy with Variable caching"),
        ('no-cache', None, "run FiPy without Variable caching"),
        ('timetests=', None, "file in which to put time spent on each test"),
        ('skfmm', None, "run FiPy using the Scikit-fmm level set solver (default)"),
        ('lsmlib', None, "run FiPy using the LSMLIB level set solver (default)"),
        ('deprecation-errors', None, "raise Exceptions for all DeprecationWarnings"),
       ]
    user_options = [_nativize_all(u) for u in user_options]

    def initialize_options(self):
        _test.initialize_options(self)

        self.all = False
        self.really_all = False
        self.examples = False
        self.modules = False
        self.viewers = False

        self.inline = False
        self.pythoncompiled = None
        self.cache = False
        self.no_cache = True
        self.Trilinos = False
        self.Pysparse = False
        self.trilinos = False
        self.pysparse = False
        self.no_pysparse = False
        self.pyamg = False
        self.pyamgx = False
        self.scipy = False
        self.petsc = False
        self.timetests = None
        self.skfmm = False
        self.lsmlib = False
        self.deprecation_errors = False

    def finalize_options(self):
        noSuiteOrModule = (self.test_suite is None
                           and self.test_module is None)

        _test.finalize_options(self)

        if noSuiteOrModule:
            # setuptools completely changed how it uses test_suite and test_args with v. 18.0
            # we do our best to keep it confused
            self.test_suite = None

        if not (self.examples or self.modules or self.viewers):
            self.all = True
        if self.all or self.really_all:
            self.examples = True
            self.modules = True
        if self.really_all:
            self.viewers = True

        # If we drop setuptools < 18.0, the remaining lines can probably be removed

        self.test_args = list(self._test_args())

        if noSuiteOrModule:
            # setuptools completely changed how it uses test_suite and test_args with v. 18.0
            # we do our best to keep it confused
            self.test_suite = "dummy"

    def _test_args(self):
        # can't seem to delegate a generator until Python 3.3

        if self.verbose:
            yield '--verbose'
        if self.test_suite:
            yield self.test_suite

        if self.viewers:
            print("*" * 60)
            print("*" + "".center(58) + "*")
            print("*" + "ATTENTION".center(58) + "*")
            print("*" + "".center(58) + "*")
            print("*" + "Some of the following tests require user interaction".center(58) + "*")
            print("*" + "".center(58) + "*")
            print("*" * 60)

            yield "fipy.viewers.testinteractive._suite"
        if self.modules:
            yield "fipy.testFiPy._suite"
        if self.examples:
            yield "examples.test._suite"

    def printPackageInfo(self):
        packages = ['python']
        versions = [sys.version.replace('\n', '| ')]

        for pkg in ['fipy', 'numpy', 'pysparse', 'scipy', 'matplotlib', 'mpi4py', 'petsc4py', 'pyamgx']:
            packages.append(pkg)

            try:
                mod = __import__(pkg)

                versions.append(mod.__version__)
            except ImportError as e:
                versions.append('not installed')

            except Exception as e:
                versions.append('version check failed: {}'.format(e))

        ## PyTrilinos
        packages.append('PyTrilinos')
        try:
            import PyTrilinos
            versions.append(PyTrilinos.version())
        except ImportError as e:
            versions.append('not installed')
        except Exception as e:
            versions.append('version check failed: {}'.format(e))

        ## Mayavi uses a non-standard approach for storing its version number.
        packages.append('mayavi')
        try:
            from mayavi.__version__ import __version__ as mayaviversion
            versions.append(mayaviversion)
        except ImportError as e:
            try:
                from enthought.mayavi.__version__ import __version__ as mayaviversion
                versions.append(mayaviversion)
            except ImportError as e:
                versions.append('not installed')
        except Exception as e:
            versions.append('version check failed: {}'.format(e))

        ## Gmsh version
        packages.append('gmsh')
        try:
            from fipy.meshes.gmshMesh import gmshVersion
            gmshversion = gmshVersion()
            if gmshversion is None:
                versions.append('not installed')
            else:
                versions.append(gmshversion)
        except Exception as e:
            versions.append('version check failed: {}'.format(e))

        packages.append("solver")
        try:
            from fipy.solvers import solver
            versions.append(solver)
        except Exception as e:
            versions.append(str(e))

        print()
        package_width = max(len(word) for word in packages)
        for package, version in zip(packages, versions):
            print("  ".join((package.ljust(package_width), version)))
        print()

    def run_tests(self):
        import sys
        if self.Trilinos or self.trilinos or self.no_pysparse:
            try:
                ## The import scipy statement is added to allow
                ## the --Trilinos tests to run without throwing a
                ## segmentation fault. This is caused by weird
                ## behavior in scipy and PyTrilinos depending on
                ## the order in which modules are imported
                try:
                    import scipy
                except:
                    pass
                import PyTrilinos
            except ImportError as a:
                print("!!! Trilinos library is not installed", file=sys.stderr)
                return

        if self.pyamgx:
            try:
                ## Unregister the function pyamgx.finalize
                ## from atexit. This prevents
                ## pyamgx from printing an error message
                ## about memory leaks and a dump of leaked memory.
                ## The memory leaks happen because
                ## the tests do not use the pyamgx solvers
                ## "cleanly", i.e., they do not use the
                ## `with` statement.
                import pyamgx
                import atexit
                if hasattr(atexit, 'unregister'):
                    atexit.unregister(pyamgx.finalize)
                else:
                    atexit._exithandlers.remove(
                        (pyamgx.finalize, (), {}))
            except ImportError as e:
                print("!!! pyamgx package is not installed", file=sys.stederr)
                return

        if self.inline:
            try:
                import weave
            except ImportError as a:
                print("!!! weave library is not installed", file=sys.stderr)
                return

        if self.pythoncompiled is not None:
            import os
            os.environ['PYTHONCOMPILED'] = self.pythoncompiled

        self.printPackageInfo()

        from pkg_resources import EntryPoint
        loader_ep = EntryPoint.parse("x="+self.test_loader)
        loader_class = loader_ep.load(require=False)

        from fipy.tools import numerix
        printoptions = numerix.get_printoptions()
        if "legacy" in printoptions:
            numerix.set_printoptions(legacy="1.13")

        if self.deprecation_errors:
            test_program_class = DeprecationErroringTestProgram
        else:
            test_program_class = unittest.TestProgram

        try:
            test_program_class(
                None, None, [unittest.__file__]+self.test_args,
                testLoader = loader_class()
                )
        except SystemExit as exitErr:
            # unittest.main(..., exit=...) not available until Python 2.7
            from fipy.tests.doctestPlus import report_skips
            report_skips()
            if self.timetests is not None:
                pass
            else:
                raise

        if "legacy" in printoptions:
            numerix.set_printoptions(legacy=printoptions["legacy"])

        if self.timetests is not None:
            from fipy.tests.doctestPlus import _DocTestTimes
            import numpy
            _DocTestTimes = numpy.rec.fromrecords(_DocTestTimes, formats='f8,S255', names='time,test')
            _DocTestTimes.sort(order=('time', 'test'))

            # Only write on proc 0.
            # Doesn't use FiPy comms because this command can
            # be run outside of FiPy (`python setup.py test`)
            try:
                from mpi4py import MPI
                procID = MPI.COMM_WORLD.rank
                barrier = MPI.COMM_WORLD.barrier
            except:
                procID = 0
                def barrier(*args):
                    pass

            if procID == 0:
                numpy.savetxt(self.timetests, _DocTestTimes[::-1],
                              delimiter='\t',
                              header="time\tmodule", comments='',
                              fmt=("%.18e", "%s"))

            barrier()

        raise exitErr



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "petscCommWrapper.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

from petsc4py import PETSc

from fipy.tools import numerix
from fipy.tools.comms.abstractCommWrapper import AbstractCommWrapper

__all__ = ["PETScCommWrapper"]

class PETScCommWrapper(AbstractCommWrapper):
    """MPI Communicator wrapper
    
    Encapsulates capabilities needed for PETSc. 
    Some capabilities are not parallel.
    """
    
    def __init__(self, petsc4py_comm=PETSc.COMM_WORLD):
        self.petsc4py_comm = petsc4py_comm
        super(PETScCommWrapper, self).__init__()
        
    @property
    def mpi4py_comm(self):
        return self.petsc4py_comm.tompi4py()
        
    def Norm2(self, vec):
        return vec.norm(norm_type=1)



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.terms.sourceTerm import SourceTerm
from fipy.tools import numerix

__all__ = ["ImplicitSourceTerm"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class ImplicitSourceTerm(SourceTerm):
    r"""

    The `ImplicitSourceTerm` represents

    .. math::

       \int_V \phi S \,dV \simeq \phi_P S_P V_P

    where :math:`S` is the `coeff` value.
    """

    def _getWeight(self, var, transientGeomCoeff=None, diffusionGeomCoeff=None):
        """
        Test for a bug due to the sign operator not being updating
        correctly.

            >>> from fipy import *
            >>> m = Grid1D(nx=1)
            >>> v = CellVariable(mesh=m, value=1.)
            >>> eq = TransientTerm() == ImplicitSourceTerm(v)
            >>> eq.solve(v, dt=1.)
            >>> print(v)
            [ 2.]
            >>> v.setValue(-1.)
            >>> eq.solve(v, dt=1.)
            >>> print(v)
            [-0.5]

        """

        coeff = self._getGeomCoeff(var)
        diagonalSign = self._getDiagonalSign(transientGeomCoeff, diffusionGeomCoeff)
        combinedSign = numerix.array(diagonalSign)[..., numerix.newaxis] * numerix.sign(coeff)

        return {'diagonal' : (combinedSign >= 0),
                'old value' : numerix.zeros(var.shape, 'd'),
                'b vector' :  -var * (combinedSign < 0),
                'new value' : numerix.zeros(var.shape, 'd')}

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()




__docformat__ = 'restructuredtext'

from fipy.solvers.petsc.petscKrylovSolver import PETScKrylovSolver

__all__ = ["LinearPCGSolver"]

class LinearPCGSolver(PETScKrylovSolver):

    """
    The `LinearPCGSolver` is an interface to the cg solver in PETSc,
    using no preconditioner by default.

    """
      
    solver = 'cg'



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

import sys

from pysparse.itsolvers import krylov

from fipy.solvers.pysparse.pysparseSolver import PysparseSolver

__all__ = ["LinearCGSSolver"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class LinearCGSSolver(PysparseSolver):

    """

    The `LinearCGSSolver` solves a linear system of equations using
    the conjugate gradient squared method (CGS), a variant of the
    biconjugate gradient method (BiCG). CGS solves linear systems with
    a general non-symmetric coefficient matrix.

    The `LinearCGSSolver` is a wrapper class for the the Pysparse_
    `itsolvers.cgs()` method.

    .. _Pysparse: http://pysparse.sourceforge.net

    """

    def __init__(self, precon=None, *args, **kwargs):
        """
        Parameters
        ----------
        precon : ~fipy.solvers.pysparse.preconditioners.preconditioner.Preconditioner, optional
        """
        import warnings
        warnings.warn("The Pysparse CGS solver may return incorrect results for some matrices", UserWarning)
        super(LinearCGSSolver, self).__init__(precon=precon, *args, **kwargs)
        self.solveFnc = krylov.cgs



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.variables.cellVariable import CellVariable
from fipy.tools.numerix import MA
from fipy.tools import numerix

__all__ = []

class _InterfaceAreaVariable(CellVariable):
    def __init__(self, distanceVar):
        """
        Creates an `_InterfaceAreaVariable` object.

        Parameters
        ----------
        distanceVar : ~fipy.variables.distanceVariable.DistanceVariable
        """
        CellVariable.__init__(self, distanceVar.mesh, hasOld=False)
        self.distanceVar = self._requires(distanceVar)

    def _calcValue(self):
        normals = numerix.array(MA.filled(self.distanceVar._cellInterfaceNormals, 0))
        areas = numerix.array(MA.filled(self.mesh._cellAreaProjections, 0))
        return numerix.sum(abs(numerix.dot(normals, areas)), axis=0)
